# Defines the full loss function, from likelihood, prior and regularization terms
# 
# Copyright (c) 2021, herculens developers and contributors

__author__ = 'aymgal', 'austinpeel'


import numpy as np
import jax.numpy as jnp
from jax import jit

from herculens.Inference.legacy.base_differentiable import Differentiable
from herculens.Util.jax_util import WaveletTransform
from herculens.Util import model_util


__all__ = ['Loss']


class Loss(Differentiable):

    # TODO: creates subclasses Likelihood, Regularization and Prior to abstract out some of the methods here

    """
    Class that manages the (auto-differentiable) loss function, defined as:
    L = - log(likelihood) - log(prior) - log(regularization)

    Note that gradient, hessian, etc. are computed in the InferenceBase class.

    Supported options are:
    - likelihood_type [default: 'chi2']: single choice among
        'chi2', 'reduced_chi2', 'l2_norm'
    - regularization_terms [default: None]: a list containing choices among
        - for a 'PIXELATED' source or lens light: 'l1_starlet_source', 'l1_battle_source', 'positivity_source'
        - for a 'PIXELATED' lens potential: 'l1_starlet_potential', 'l1_battle_potential', 'positivity_potential'
    - prior_terms [default: None]: a list containing choices among
        'uniform', 'gaussian'
    """

    _supported_ll = ('chi2', 'reduced_chi2', 'l2_norm')
    _supported_regul_source = (
                               'l1_starlet_source', 
                               'l1_battle_source', 
                               'positivity_source'
                               )
    _supported_regul_lens_mass = (
                                  'l1_starlet_potential', 
                                  'l1_battle_potential', 
                                  'positivity_potential', 
                                  'negativity_potential', 
                                  'positivity_convergence',

                                  'analytical_potential',  # TEST: by default, regularize the penultimate (index -2) lens profile with the last one (index -1)
                                  )
    _supported_regul_lens_light = ('l1_starlet_lens_light', 'l1_battle_lens_light', 'positivity_lens_light')
    _supported_prior = ('uniform', 'gaussian')

    def __init__(self, data, image_class, param_class, 
                 likelihood_type='chi2', likelihood_mask=None, 
                 regularization_terms=None, regularization_strengths=None, 
                 regularization_weights=None, regularization_masks=None,
                 prior_terms=None, starlet_second_gen=False, index_analytical_potential=None):
        self._data  = data
        self._image = image_class
        self._param = param_class
        
        self._check_choices(likelihood_type, prior_terms, 
                            regularization_terms, regularization_strengths, 
                            regularization_weights, regularization_masks)
        self._init_likelihood(likelihood_type, likelihood_mask)
        self._init_regularizations(regularization_terms, 
                                   regularization_strengths, 
                                   regularization_weights, 
                                   regularization_masks, 
                                   starlet_second_gen, index_analytical_potential)
        self._init_priors(prior_terms)

    def _func(self, args):
        """negative log(likelihood*prior*regularization)"""
        kwargs = self._param.args2kwargs(args)
        model = self._image.model(**kwargs, k_lens=self._regul_k_lens)
        neg_log_ll  = - self.log_likelihood(model)
        neg_log_reg = - self.log_regularization(kwargs)
        neg_log_p   = - self.log_prior(args)
        neg_log = neg_log_ll + neg_log_reg + neg_log_p
        neg_log /= self._global_norm  # to keep loss magnitude in acceptable range
        return jnp.nan_to_num(neg_log, nan=1e15, posinf=1e15, neginf=1e15)

    @property
    def likelihood_num_data_points(self):
        return self._ll_num_data_points

    @property
    def likelihood_mask(self):
        return self._ll_mask

    @property
    def data(self):
        return self._data

    def _check_choices(self, likelihood_type, prior_terms, 
                       regularization_terms, regularization_strengths, 
                       regularization_weights, regularization_masks):
        if likelihood_type not in self._supported_ll:
            raise ValueError(f"Likelihood term '{likelihood_type}' is not supported")
        if prior_terms is not None:
            for term in prior_terms:
                if term not in self._supported_prior:
                    raise ValueError(f"Prior term '{term}' is not supported")
        if regularization_terms is not None:
            if len(regularization_terms) != len(regularization_strengths):
                raise ValueError(f"There should be at least one choice of "
                                 "regularization strength per regularization term.")
            if likelihood_type in ['chi2', 'reduced_chi2']:
                UserWarning(f"Likelihood type is '{likelihood_type}', which might "
                            "cause issues with some regularization choices")
            for term in regularization_terms:
                if term not in (self._supported_regul_source + 
                                self._supported_regul_lens_mass +
                                self._supported_regul_lens_light):
                    raise ValueError(f"Regularization term '{term}' is not supported")
                # TODO: if any regularization terms are not dependent on PIXELATED profiles
                # need to update these checks below
                if (term in self._supported_regul_source and 
                    'PIXELATED' not in self._image.SourceModel.profile_type_list):
                    raise ValueError(f"Regularization term '{term}' is only "
                                     "compatible with a 'PIXELATED' source light profile")
                
                if (term in self._supported_regul_lens_mass and 
                    'PIXELATED' not in self._image.MassModel.profile_type_list):
                    raise ValueError(f"Regularization term '{term}' is only "
                                     "compatible with a 'PIXELATED' lens profile")

                if (term in self._supported_regul_lens_light and 
                    'PIXELATED' not in self._image.LensLightModel.profile_type_list):
                    raise ValueError(f"Regularization term '{term}' is only "
                                     "compatible with a 'PIXELATED' lens profile")

    def _init_likelihood(self, likelihood_type, likelihood_mask):
        if likelihood_mask is None:
            self._ll_mask = np.ones_like(self._data)
        else:
            self._ll_mask = likelihood_mask.astype(float)
        self._ll_num_data_points = np.count_nonzero(self._ll_mask)
        if likelihood_type == 'chi2':
            self.log_likelihood = self.log_likelihood_chi2
            self._global_norm = 1.
        elif likelihood_type == 'reduced_chi2':
            self.log_likelihood = self.log_likelihood_chi2
            self._global_norm = 0.5 * self.likelihood_num_data_points
        elif likelihood_type == 'l2_norm':
            self.log_likelihood = self.log_likelihood_l2
            # here the global norm is such that l2_norm has same order of magnitude as a chi2
            self._global_norm = 1.0 # 0.5 * self._image.Grid.num_pixel * np.mean(self._image.Noise.C_D)

    def _init_regularizations(self, regularization_terms, regularization_strengths, 
                              regularization_weights, regularization_masks,
                              starlet_second_gen, index_analytical_potential):

        self._regul_k_lens = None  # TEMPORARY

        if regularization_terms is None:
            self.log_regularization = lambda kwargs: 0.  # no regularization
            return

        if regularization_masks is None:
            regularization_masks = [None]*len(regularization_terms)

        # TODO: implement regularization_weights for source regularization as well (for now it's only potential)
        i = 0
        regularization_weights_fix = []
        for term in regularization_terms:
            if 'potential' in term:
                regularization_weights_fix.append(regularization_weights[i])
                i += 1
            else:
                # TEMPORARY: just to populate weights for regularization terms other than potential
                # waiting for the source and lens light weights to be handled as well.
                regularization_weights_fix.append(None)

        self._idx_pix_src = self._image.SourceModel.pixelated_index
        self._idx_pix_pot = self._image.MassModel.pixelated_index
        self._idx_pix_ll  = self._image.LensLightModel.pixelated_index

        regul_func_list = []
        for term, strength, weights, mask in zip(regularization_terms, 
                                           regularization_strengths, 
                                           regularization_weights_fix,
                                           regularization_masks):
            # add the log-regularization function to the list
            regul_func_list.append(getattr(self, '_log_regul_'+term))

            if term == 'l1_starlet_source':
                n_pix_src = min(*self._image.SourceModel.pixelated_shape)
                n_scales = int(np.log2(n_pix_src))  # maximum allowed number of scales
                self._starlet_src = WaveletTransform(n_scales, wavelet_type='starlet',
                                                     second_gen=starlet_second_gen)
                wavelet_norms = self._starlet_src.scale_norms[:-1]  # ignore coarsest scale
                self._st_src_norms = jnp.expand_dims(wavelet_norms, (1, 2))
                if isinstance(strength, (int, float)):
                    self._st_src_lambda = self._st_src_lambda_hf = float(strength)
                elif isinstance(strength, (tuple, list)):
                    if len(strength) > 2:
                        raise ValueError("You can only specify two starlet regularization "
                                         "strength values at maximum")
                    self._st_src_lambda_hf = float(strength[0])
                    self._st_src_lambda = float(strength[1])

            elif term == 'l1_starlet_lens_light':
                n_pix_ll = min(*self._image.LensLightModel.pixelated_shape)
                n_scales = int(np.log2(n_pix_ll))  # maximum allowed number of scales
                self._starlet_ll = WaveletTransform(n_scales, wavelet_type='starlet',
                                                    second_gen=starlet_second_gen)
                wavelet_norms = self._starlet_ll.scale_norms[:-1]  # ignore coarsest scale
                self._st_ll_norms = jnp.expand_dims(wavelet_norms, (1, 2))
                if isinstance(strength, (int, float)):
                    self._st_ll_lambda = float(strength)
                    self._st_ll_lambda_hf = float(strength)
                elif isinstance(strength, (tuple, list)):
                    if len(strength) > 2:
                        raise ValueError("You can only specify two starlet regularization "
                                         "strength values at maximum")
                    self._st_ll_lambda_hf = float(strength[0])
                    self._st_ll_lambda = float(strength[1])

            elif term == 'l1_battle_source':
                n_scales = 1  # maximum allowed number of scales
                self._battle_src = WaveletTransform(n_scales, wavelet_type='battle-lemarie-3')
                self._bl_src_norm = self._battle_src.scale_norms[0]  # consider only first scale
                if isinstance(strength, (tuple, list)):
                    raise ValueError("You can only specify one regularization "
                                     "strength for Battle-Lemarie regularization")
                self._bl_src_lambda = float(strength)

            elif term == 'l1_battle_lens_light':
                n_scales = 1  # maximum allowed number of scales
                self._battle_ll = WaveletTransform(n_scales, wavelet_type='battle-lemarie-3')
                self._bl_ll_norm = self._battle_ll.scale_norms[0]  # consider only first scale
                if isinstance(strength, (tuple, list)):
                    raise ValueError("You can only specify one regularization "
                                     "strength for Battle-Lemarie regularization")
                self._bl_ll_lambda = float(strength)

            elif term == 'positivity_source':
                if isinstance(strength, (tuple, list)):
                    raise ValueError("You can only specify one regularization "
                                     "strength for positivity constraint")
                self._pos_src_lambda = float(strength)

            elif term == 'positivity_lens_light':
                if isinstance(strength, (tuple, list)):
                    raise ValueError("You can only specify one regularization "
                                     "strength for positivity constraint")
                self._pos_ll_lambda = float(strength)

            elif term == 'l1_starlet_potential':
                n_pix_pot = min(*self._image.MassModel.pixelated_shape)
                n_scales = int(np.log2(n_pix_pot))  # maximum allowed number of scales
                self._starlet_pot = WaveletTransform(n_scales, wavelet_type='starlet',
                                                     second_gen=starlet_second_gen)
                wavelet_norms = self._starlet_pot.scale_norms[:-1]  # ignore coarsest scale
                # self._st_pot_norms = jnp.expand_dims(wavelet_norms, (1, 2))
                if weights.shape[0] != n_scales+1:
                    raise ValueError(f"The weights do not contain enough wavelet scales"
                                     f" (should be {n_scales+1} inc. coarsest).")
                self._st_pot_weigths = weights
                if isinstance(strength, (int, float)):
                    self._st_pot_lambda = float(strength)
                    self._st_pot_lambda_hf = float(strength)
                elif isinstance(strength, (tuple, list)):
                    if len(strength) > 2:
                        raise ValueError("You can only specify two starlet regularization "
                                         "strength values at maximum")
                    self._st_pot_lambda_hf = float(strength[0])
                    self._st_pot_lambda = float(strength[1])

            elif term == 'l1_battle_potential':
                n_scales = 1  # maximum allowed number of scales
                self._battle_pot = WaveletTransform(n_scales, wavelet_type='battle-lemarie-3')
                # self._bl_pot_norm = self._battle_pot.scale_norms[0]  # consider only first scale
                if weights.shape[0] != n_scales+1:
                    raise ValueError(f"The weights do not contain enogh wavelet scales"
                                     f" (should be {n_scales+1} inc. coarsest).")
                self._bl_pot_weigths = weights
                if isinstance(strength, (tuple, list)):
                    raise ValueError("You can only specify one regularization "
                                     "strength for Battle-Lemarie regularization")
                self._bl_pot_lambda = float(strength)

            elif term == 'positivity_potential':
                if isinstance(strength, (tuple, list)):
                    raise ValueError("You can only specify one regularization "
                                     "strength for positivity constraint")
                self._pos_pot_lambda = float(strength)

            elif term == 'negativity_potential':
                if isinstance(strength, (tuple, list)):
                    raise ValueError("You can only specify one regularization "
                                     "strength for positivity constraint")
                self._neg_pot_lambda = float(strength)

            elif term == 'positivity_convergence':
                if isinstance(strength, (tuple, list)):
                    raise ValueError("You can only specify one regularization "
                                     "strength for positivity constraint")
                self._pos_conv_lambda = float(strength)
                self._x_lens, self._y_lens = self._image.Grid.model_pixel_coordinates('lens')


            elif term == 'analytical_potential':
                if index_analytical_potential is None:
                    raise ValueError("For analytical potential regularization, a `index_analytical_potential` is required.")
                self._idx_ana_pot = index_analytical_potential
                self._regul_k_lens = tuple([True if i != self._idx_ana_pot else False for i in range(len(self._image.MassModel.profile_type_list))])
                self._weigths = weights
                self._lambda  = float(strength)
                self._mask    = mask
                self._x_lens, self._y_lens = self._image.Grid.model_pixel_coordinates('lens')


        # build the composite function (sum of regularization terms)
        self.log_regularization = lambda kw: sum([func(kw) for func in regul_func_list])

    def _init_priors(self, prior_terms):
        if prior_terms is None:
            self.log_prior = lambda args: 0.
            return
        if prior_terms == ['uniform']:
            self.log_prior = self._param.log_prior_uniform
        elif prior_terms == ['gaussian']:
            self.log_prior = self._param.log_prior_gaussian
        elif 'gaussian' in prior_terms and 'uniform' in prior_terms:
            self.log_prior = self._param.log_prior

    # def log_likelihood_gaussian(self, model):
    #     C_D = self._image.Noise.C_D_model(model)
    #     det_C_D = jnp.prod(noise_var)  # knowing that C_D is diagonal
    #     #print("det_C_D", det_C_D)
    #     Z_D = np.sqrt( (2*np.pi)**self.likelihood_num_data_points * det_C_D )  # Eq. 24 from Vegetti & Koopmans 2009
    #     chi2 = - 0.5 * jnp.sum( (self._data - model)**2 * self.likelihood_mask / C_D )
    #     return jnp.log(Z_D) + chi2

    def log_likelihood_chi2(self, model):
        noise_var = self._image.Noise.C_D_model(model)
        # noise_var = self._image.Noise.C_D
        residuals = (self._data - model) * self.likelihood_mask
        return - 0.5 * jnp.sum(residuals**2 / noise_var)

    def log_likelihood_l2(self, model):
        # TODO: check that mask here does not mess up with the balance between l2-norm and wavelet regularization
        residuals = (self._data - model) * self.likelihood_mask
        return - 0.5 * jnp.sum(residuals**2)

    def _log_regul_l1_starlet_source(self, kwargs):
        model = self._image.model(**kwargs)
        noise_map = jnp.sqrt(self._image.Noise.C_D_model(model))  # TODO: do not take into account shot noise from lens light
        noise_level = jnp.mean(noise_map[self.likelihood_mask == 1])

        source_model = kwargs['kwargs_source'][self._idx_pix_src]['pixels']
        st = self._starlet_src.decompose(source_model)[:-1]  # ignore coarsest scale
        st_weighted_l1_hf = jnp.sum(self._st_src_norms[0] * noise_level * jnp.abs(st[0]))  # first scale (i.e. high frequencies)
        st_weighted_l1 = jnp.sum(self._st_src_norms[1:] * noise_level * jnp.abs(st[1:]))  # other scales
        return - (self._st_src_lambda_hf * st_weighted_l1_hf + self._st_src_lambda * st_weighted_l1)

    def _log_regul_l1_starlet_lens_light(self, kwargs):
        # TODO: generalise this for Poisson noise! but then the noise needs to be properly propagated to source plane
        noise_map = np.sqrt(self._image.Noise.C_D)

        # TEST reweight the noise map based on lensed source model
        #lensed_source_model = self._image.source_surface_brightness(kwargs['kwargs_source'], 
        #                                                            kwargs_lens=kwargs['kwargs_lens'],
        #                                                            de_lensed=False, unconvolved=True)
        #noise_level = noise_map # + lensed_source_model**3
        noise_level = np.mean(noise_map[self.likelihood_mask == 1])
        # end TEST

        model = kwargs['kwargs_lens_light'][self._idx_pix_ll]['pixels']
        st = self._starlet_ll.decompose(model)[:-1]  # ignore coarsest scale
        st_weighted_l1_hf = jnp.sum(self._st_ll_norms[0] * noise_level * jnp.abs(st[0]))  # first scale (i.e. high frequencies)
        st_weighted_l1 = jnp.sum(self._st_ll_norms[1:] * noise_level * jnp.abs(st[1:]))  # other scales
        return - (self._st_ll_lambda_hf * st_weighted_l1_hf + self._st_ll_lambda * st_weighted_l1)

    def _log_regul_l1_starlet_potential(self, kwargs):
        weights = self._st_pot_weigths
        psi_model = kwargs['kwargs_lens'][self._idx_pix_pot]['pixels']
        st = self._starlet_pot.decompose(psi_model)
        st_weighted_l1_hf = jnp.sum(jnp.abs(weights[0] * st[0]))  # first scale (i.e. high frequencies)
        st_weighted_l1 = jnp.sum(jnp.abs(weights[1:-1] * st[1:-1]))  # other scales (except coarsest)
        return - (self._st_pot_lambda_hf * st_weighted_l1_hf + self._st_pot_lambda * st_weighted_l1)

    def _log_regul_l1_battle_source(self, kwargs):
        model = self._image.model(**kwargs)
        noise_map = jnp.sqrt(self._image.Noise.C_D_model(model))  # TODO: do not take into account shot noise from lens light
        noise_level = jnp.mean(noise_map[self.likelihood_mask == 1])
        source_model = kwargs['kwargs_source'][self._idx_pix_src]['pixels']
        bl = self._battle_src.decompose(source_model)[0]  # consider only first scale
        bl_weighted_l1 = jnp.sum(self._bl_src_norm * noise_level * jnp.abs(bl))
        return - self._bl_src_lambda * bl_weighted_l1

    def _log_regul_l1_battle_lens_light(self, kwargs):
        # TODO: generalise this for Poisson noise! but then the noise needs to be properly propagated to source plane
        noise_map = np.sqrt(self._image.Noise.C_D)
        noise_level = np.mean(noise_map[self.likelihood_mask == 1])
        #noise_level = noise_map
        model = kwargs['kwargs_lens_light'][self._idx_pix_ll]['pixels']
        bl = self._battle_ll.decompose(model)[0]  # consider only first scale
        bl_weighted_l1 = jnp.sum(self._bl_ll_norm * noise_level * jnp.abs(bl))
        return - self._bl_ll_lambda * bl_weighted_l1

    def _log_regul_l1_battle_potential(self, kwargs):
        weights = self._bl_pot_weigths
        psi_model = kwargs['kwargs_lens'][self._idx_pix_pot]['pixels']
        bl = self._battle_pot.decompose(psi_model)
        bl_weighted_l1 = jnp.sum(jnp.abs(weights[0] * bl[0]))  # only first BL scale
        return - self._bl_pot_lambda * bl_weighted_l1

    def _log_regul_positivity_source(self, kwargs):
        source_model = kwargs['kwargs_source'][self._idx_pix_src]['pixels']
        return - self._pos_src_lambda * jnp.abs(jnp.sum(jnp.minimum(0., source_model)))

    def _log_regul_positivity_lens_light(self, kwargs):
        model = kwargs['kwargs_lens_light'][self._idx_pix_ll]['pixels']
        return - self._pos_ll_lambda * jnp.abs(jnp.sum(jnp.minimum(0., model)))

    def _log_regul_positivity_potential(self, kwargs):
        psi_model = kwargs['kwargs_lens'][self._idx_pix_pot]['pixels']
        return - self._pos_pot_lambda * jnp.abs(jnp.sum(jnp.minimum(0., psi_model)))

    def _log_regul_negativity_potential(self, kwargs):
        psi_model = kwargs['kwargs_lens'][self._idx_pix_pot]['pixels']
        return - self._neg_pot_lambda * jnp.abs(jnp.sum(jnp.maximum(0., psi_model)))

    def _log_regul_positivity_convergence(self, kwargs):
        kappa_model = self._image.MassModel.kappa(self._x_lens, 
                                                  self._y_lens,
                                                  kwargs['kwargs_lens'], 
                                                  k=self._idx_pix_pot)
        return - self._pos_conv_lambda * jnp.abs(jnp.sum(jnp.minimum(0., kappa_model)))

    def _log_regul_analytical_potential(self, kwargs):
        psi_model = kwargs['kwargs_lens'][self._idx_pix_pot]['pixels']
        target_model = self._image.MassModel.potential(self._x_lens, self._y_lens, 
                                                       kwargs['kwargs_lens'],
                                                       k=self._idx_ana_pot)
        return - self._lambda * jnp.sum(self._mask * self._weigths * (psi_model - target_model)**2)
        # or similar to Tagore & Keeton 2014 (does not seem to work tho)
        #return - self._lambda * (jnp.sum(self._mask * self._weigths * psi_model / target_model) - jnp.sum(self._mask) * jnp.mean(psi_model / target_model))



# Defines the model of a strong lens
#
# Copyright (c) 2023, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the ImSim module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'austinpeel', 'aymgal'

import copy
import numpy as np
import jax.numpy as jnp
from functools import partial
from jax import jit
from jax import random

from herculens.LensImage.lens_image_multiplane import MPLensImage
from herculens.MassModel.mass_model_multiplane import MPMassModel
from herculens.LightModel.light_model_multiplane import MPLightModel
# from herculens.LensImage.Numerics.numerics import Numerics
from herculens.LensImage.lensing_operator import LensingOperator


__all__ = [
    'LensImage', 
    # 'LensImage3D',  # TODO: to implement
]


class LensImage(MPLensImage):
    """Generate lensed images from source light, lens mass/light, and point source models."""

    def __init__(self, 
                 grid_class, 
                 psf_class,
                 noise_class=None, 
                 lens_mass_model_class=None,
                 source_model_class=None, 
                 lens_light_model_class=None,
                 point_source_model_class=None, 
                 source_arc_mask=None, 
                 source_grid_scale=1.,
                 kwargs_numerics=None,
                 kwargs_lens_equation_solver=None,
                 conjugate_points=None,
                 deflection_scaling_convention='standard',
                 deflection_scaling=1.,
                 extra_surface_brightness=None):
        """
        WIP
        """
        # mp_mass_model_list = [lens_mass_model_class.func_list]  # single mass plane
        mp_mass_model_class = MPMassModel(
            [lens_mass_model_class],
            deflection_scaling_convention=deflection_scaling_convention,
        )
        # mp_light_model_list = []
        # light_model_kwargs = []
        # if lens_light_model_class is not None:
        #     # first light plane
        #     mp_light_model_list.append(lens_light_model_class.func_list)
        #     light_model_kwargs.append({'kwargs_pixelated': lens_light_model_class.pixel_grid_settings})
        # else:
        #     mp_light_model_list.append([])
        #     light_model_kwargs.append({})
        # if source_model_class is not None:
        #     # second light plane
        #     mp_light_model_list.append(source_model_class.func_list)
        #     light_model_kwargs.append({'kwargs_pixelated': source_model_class.pixel_grid_settings})
        # else:
        #     mp_light_model_list.append([])
        #     light_model_kwargs.append({})
        mp_light_model_class = MPLightModel([lens_light_model_class, source_model_class])
        super().__init__(
            grid_class,
            psf_class,
            noise_class,
            mp_mass_model_class,
            mp_light_model_class,
            source_arc_masks=[None, source_arc_mask],  # no mask for lens light plane
            source_grid_scale=[None, source_grid_scale],  # no grid scale for lens light
            conjugate_points=[conjugate_points],
            kwargs_numerics=kwargs_numerics
        )
        self.eta_flat_fixed = np.array([deflection_scaling])  # this is fixed in single lens plane mode

        # the following attributes are to mimick the original implementation of LensImage
        self.MassModel = lens_mass_model_class
        self.LensLightModel = lens_light_model_class
        self.SourceModel = source_model_class  # in single plane there is 1 source model
        if point_source_model_class is not None:
            raise NotImplementedError("Point source are not implemented in this class yet")
        else:
            from herculens.PointSourceModel.point_source_model import PointSourceModel
            point_source_model_class = PointSourceModel(
                point_source_type_list=[], mass_model=lens_mass_model_class,
            )
        self.PointSourceModel = point_source_model_class

        # WIP: extra surface brightness image to add to the model (should be already convolved)
        self.add_fixed_light = extra_surface_brightness is not None
        self._extra_light_model = extra_surface_brightness

    # WIP
    def set_static_model_grid(self):
        ra_grid_img, dec_grid_img = self.ImageNumerics.coordinates_evaluate
        for i in range(self.MPMassModel.number_mass_planes):
            for j in range(self.MPMassModel.mass_models[i]._num_func):
                profile = self.MPMassModel.mass_models[i].func_list[j]
                if hasattr(profile, 'set_eval_coord_grid'):
                    profile.set_eval_coord_grid(ra_grid_img, dec_grid_img)
                    print(f"Set static coordinate grid for profile {j} in mass plane {i}")
            

    @partial(jit, static_argnums=(0, 5, 6, 7, 8, 9, 10, 11, 12, 13))
    def model(self, kwargs_lens=None, kwargs_source=None, kwargs_lens_light=None,
              kwargs_point_source=None, unconvolved=False, supersampled=False,
              source_add=True, lens_light_add=True, point_source_add=True,
              k_lens=None, k_source=None, k_lens_light=None, k_point_source=None):
        """
        Create the 2D model image from parameter values.
        Note: due to JIT compilation, the first call to this method will be slower.

        :param kwargs_lens: list of keyword arguments corresponding to the superposition of different lens profiles
        :param kwargs_source: list of keyword arguments corresponding to the superposition of different source light profiles
        :param kwargs_lens_light: list of keyword arguments corresponding to different lens light surface brightness profiles
        :param kwargs_point_source: keyword arguments corresponding to "other" parameters, such as external shear and
                                    point source image positions
        :param unconvolved: if True: returns the unconvolved light distribution (prefect seeing)
        :param supersampled: if True, returns the model on the higher resolution grid (WARNING: no convolution nor normalization is performed in this case!)
        :param source_add: if True, compute source, otherwise without
        :param lens_light_add: if True, compute lens light, otherwise without
        :param point_source_add: if True, compute point source multiple images, otherwise without
        :param k_lens: list of bool or list of int to select which lens mass profiles to include
        :param k_source: list of bool or list of int to select which source profiles to include
        :param k_lens_light: list of bool or list of int to select which lens light profiles to include
        :param k_point_source: list of bool or list of int to select which point sources to include
        :return: 2d array of surface brightness pixels of the simulation
        """
        if kwargs_point_source is not None:
            raise NotImplementedError
        if k_lens is None:
            k_mass = None
        else:
            raise NotImplementedError
        if k_lens_light is None and k_source is None:
            k_light = None
        else:
            raise NotImplementedError
        model = super().model(
            eta_flat=self.eta_flat_fixed,
            kwargs_mass=[kwargs_lens],
            kwargs_light=[kwargs_lens_light, kwargs_source],
            supersampled=supersampled,
            unconvolved=unconvolved,
            k_mass=k_mass,
            k_light=k_light,
            k_planes=-1,
            return_pixel_scale=False,
        )
        if self.add_fixed_light:
            model += self._extra_light_model
        return model

    # def source_surface_brightness(self, kwargs_source, kwargs_lens=None,
    #                               unconvolved=False, supersampled=False,
    #                               de_lensed=False, k=None, k_lens=None):
    #     """

    #     computes the source surface brightness distribution

    #     :param kwargs_source: list of keyword arguments corresponding to the superposition of different source light profiles
    #     :param kwargs_lens: list of keyword arguments corresponding to the superposition of different lens profiles.
    #     When using an adaptive source pixel grid, kwargs_lens is required even if de_lensed=True.
    #     :param kwargs_extinction: list of keyword arguments of extinction model
    #     :param unconvolved: if True: returns the unconvolved light distribution (prefect seeing)
    #     :param de_lensed: if True: returns the un-lensed source surface brightness profile, otherwise the lensed.
    #     :param k: list of bool or list of int to select which source profiles to include
    #     :param k_lens: list of bool or list of int to select which lens mass profiles to include
    #     :return: 2d array of surface brightness pixels
    #     """
    #     if len(self.SourceModel.profile_type_list) == 0:
    #         return jnp.zeros(self.Grid.num_pixel_axes)
    #     x_grid_img, y_grid_img = self.ImageNumerics.coordinates_evaluate
    #     source_light = self.eval_source_surface_brightness(
    #         x_grid_img, y_grid_img,
    #         kwargs_source, kwargs_lens=kwargs_lens,
    #         k=k, k_lens=k_lens, de_lensed=de_lensed,
    #     )
    #     if not supersampled:
    #         source_light = self.ImageNumerics.re_size_convolve(
    #             source_light, unconvolved=unconvolved)
    #     return source_light

    # def lens_surface_brightness(self, kwargs_lens_light, unconvolved=False,
    #                             supersampled=False, k=None):
    #     """

    #     computes the lens surface brightness distribution

    #     :param kwargs_lens_light: list of keyword arguments corresponding to different lens light surface brightness profiles
    #     :param unconvolved: if True, returns unconvolved surface brightness (perfect seeing), otherwise convolved with PSF kernel
    #     :param k: list of bool or list of int to select which model profiles to include
    #     :return: 2d array of surface brightness pixels
    #     """
    #     x_grid_img, y_grid_img = self.ImageNumerics.coordinates_evaluate
    #     lens_light = self.LensLightModel.surface_brightness(x_grid_img, y_grid_img, kwargs_lens_light, k=k)
    #     if not supersampled:
    #         lens_light = self.ImageNumerics.re_size_convolve(
    #             lens_light, unconvolved=unconvolved)
    #     return lens_light

    # def point_source_image(self, kwargs_point_source, kwargs_lens,
    #                        kwargs_solver, k=None):
    #     """Compute PSF-convolved point sources rendered on the image plane.

    #     :param kwargs_point_source: list of keyword arguments corresponding to the point sources
    #     :param kwargs_lens: list of keyword arguments corresponding to the superposition of different lens profiles
    #     :param k: list of bool or list of int to select which point sources to include
    #     :return: 2d array at the image plane resolution of all multiple images of the point sources
    #     """
    #     result = jnp.zeros((self.Grid.num_pixel_axes))
    #     if self.PointSourceModel is None:
    #         return result
    #     theta_x, theta_y, amplitude = self.PointSourceModel.get_multiple_images(
    #         kwargs_point_source, kwargs_lens=kwargs_lens, kwargs_solver=kwargs_solver, 
    #         k=k, with_amplitude=True, zero_amp_duplicates=True
    #     )
    #     for i in range(len(theta_x)):
    #         result += self.ImageNumerics.render_point_sources(
    #             theta_x[i], theta_y[i], amplitude[i]
    #         )
    #     return result

    # def adapt_source_coordinates(self, kwargs_lens, k_lens=None, return_plt_extent=False):
    #     # NOTE: despite the keyword name, this does *NOT* return the extent
    #     # as expected from pyplot routines! # TODO: fix this
    #     if k_lens is not None:
    #         raise NotImplementedError("k_lens != None not implemented in new LensImage class.")
    #     ra_grid_img, dec_grid_img = self.ImageNumerics.coordinates_evaluate
    #     kwargs_mass = [kwargs_lens]  # wraps the single mass model for multiplane conventions
    #     ra_grid_planes, dec_grid_planes = self.MPMassModel.ray_shooting(
    #         ra_grid_img,
    #         dec_grid_img,
    #         eta_flat,
    #         kwargs_mass
    #     )
    #     x_coord_list, y_coord_list, extent_list = super().adapt_source_coordinates(
    #         ra_grid_planes, 
    #         dec_grid_planes,
    #         force=False,
    #         npix_src=100,  # not used when force=False
    #         source_grid_scale=1.0  # TODO: implement this
    #     )
    #     # extract the source as the last light model in multiplane conventions
    #     x_coord, y_coord, extent = x_coord_list[-1], y_coord_list[-1], extent_list[-1]
    #     return (x_coord, y_coord, extent)

    def get_source_coordinates(self, kwargs_lens, k_lens=None, return_plt_extent=False):
        """
        WARNING: this does not return the same array shape as the super() class!
        This function returns 2D 'mesh-gridded' arrays so they can be used by
        the plotting routines, and is consistent with the output of the 
        implementation in the original LensImage class.
        """
        if k_lens is not None:
            raise NotImplementedError("k_lens != None not implemented in new LensImage class.")
        # NOTE: despite the keyword name, this does *NOT* return the extent
        # as expected from pyplot routines! # TODO: fix this 
        kwargs_mass = [kwargs_lens]  # wraps the single mass model for multiplane conventions
        x_coord_list, y_coord_list, extent_list = super().get_source_coordinates(
            self.eta_flat_fixed,
            kwargs_mass,
            force=False,
            npix_src=100,  # not used when force=False
            source_grid_scale=self._source_grid_scale[-1], 
        )
        # extract the source as the last light model in multiplane conventions
        x_coord, y_coord, extent = x_coord_list[-1], y_coord_list[-1], extent_list[-1]
        # generate the 2D coordinates grid
        x_grid, y_grid = np.meshgrid(x_coord, y_coord)
        return (x_grid, y_grid, extent)

    def eval_source_surface_brightness(self, x, y, kwargs_source, kwargs_lens=None, 
                                       k=None, k_lens=None, de_lensed=False):
        # pixels_x_coord, pixels_y_coord, _ = self.adapt_source_coordinates(
        #     kwargs_lens, k_lens=k_lens,
        # )

        if k_lens is not None:
            raise NotImplementedError("k_lens != None not implemented in new LensImage class.")

        ra_grid_img, dec_grid_img = self.ImageNumerics.coordinates_evaluate
        kwargs_mass = [kwargs_lens]  # wraps the single mass model for multiplane conventions
        ra_grid_planes, dec_grid_planes = self.MPMassModel.ray_shooting(
            ra_grid_img,
            dec_grid_img,
            self.eta_flat_fixed,
            kwargs_mass
        )
        x_coord_list, y_coord_list, _ = super().adapt_source_coordinates(
            ra_grid_planes, 
            dec_grid_planes,
            force=False,
            npix_src=100,  # not used when force=False
            source_grid_scale=self._source_grid_scale[-1], 
        )
        # extract the source as the last light model in multiplane conventions
        pixels_x_coord, pixels_y_coord = x_coord_list[-1], y_coord_list[-1]

        if de_lensed is True:
            source_light = self.SourceModel.surface_brightness(x, y, kwargs_source, k=k,
                                                               pixels_x_coord=pixels_x_coord, pixels_y_coord=pixels_y_coord)
        else:
            kwargs_mass = [kwargs_lens]
            x_grid_src, y_grid_src = self.MPMassModel.ray_shooting(x, y, kwargs_mass, k=k_lens)
            source_light = self.SourceModel.surface_brightness(x_grid_src, y_grid_src, kwargs_source, k=k,
                                                               pixels_x_coord=pixels_x_coord, pixels_y_coord=pixels_y_coord)
        return source_light

    def normalized_residuals(self, data, model, kwargs_noise=None, mask=None):
        """
        compute the map of normalized residuals,
        given the data and the model image
        """
        if kwargs_noise is None:
            kwargs_noise = {}
        if mask is None:
            mask = np.ones(self.Grid.num_pixel_axes)
        noise_var = self.Noise.C_D_model(model, **kwargs_noise)
        noise = np.sqrt(noise_var)
        norm_res_model = (data - model) / noise * mask
        norm_res_tot = norm_res_model
        if mask is not None:
            # outside the mask just add pure data
            norm_res_tot += (data / noise) * (1. - mask)
        # make sure there is no NaN or infinite values
        norm_res_model = np.where(np.isfinite(norm_res_model), norm_res_model, 0.)
        norm_res_tot = np.where(np.isfinite(norm_res_tot), norm_res_tot, 0.)
        return norm_res_model, norm_res_tot

    def reduced_chi2(self, data, model, kwargs_noise=None, mask=None):
        """
        compute the reduced chi2 of the data given the model
        """
        if mask is None:
            mask = np.ones(self.Grid.num_pixel_axes)
        norm_res, _ = self.normalized_residuals(data, model, kwargs_noise=kwargs_noise, mask=mask)
        num_data_points = np.sum(mask)
        return np.sum(norm_res**2) / num_data_points
        
    @partial(jit, static_argnums=(0, 2))
    def trace_conjugate_points(self, kwargs_mass, k_mass=None):
        '''
        Helper function that can be used to ray-trace the list of conjugate points
        provided to the class on initialization to their corresponding source planes.
        '''
        return super().trace_conjugate_points(
            self.eta_flat_fixed,
            kwargs_mass,
            N=1,
            k_mass=k_mass,
        )
    


# Gradient-based sensitivity mapping
# 
# Copyright (c) 2021, herculens developers and contributors

__author__ = 'aymgal'


import copy
import time
import numpy as np
from skimage import feature
from functools import partial

from jax import grad
from jax import jit
import jax.numpy as jnp

from herculens.LightModel.light_model import LightModel
from herculens.MassModel.mass_model import MassModel
from herculens.LensImage.lens_image import LensImage
from herculens.Inference.legacy.parameters import Parameters
from herculens.Inference.legacy.loss import Loss
from herculens.Inference.legacy.optimization import Optimizer
# from herculens.Inference.legacy.covariance import FisherCovariance
from herculens.Util import util


__all__ = ['MassSensitivityMapping']



class MassSensitivityMapping(object):

    _MODELS_SUPPORTED = [
        'POINT_MASS',       # point mass (analytical)
        'SIS',              # Singuler Isothermal Sphere (analytical)
        'PIXELATED_DIRAC',  # single potential 'pixel'
        # TODO: add NFW and pseudo-Jaffe profiles
    ]

    def __init__(self, data, macro_lens_image, macro_parameters, macro_loss, 
                 halo_profile='SIS', fix_macro_parameters=True, 
                 kwargs_numerics=None, verbose=False):
        self.data = data
        self.m_lens_image = macro_lens_image
        self.m_param = macro_parameters
        self.m_loss = macro_loss
        self.kwargs_numerics = kwargs_numerics
        self.fix_macro = fix_macro_parameters
        self.halo_profile = halo_profile
        self.verbose = verbose

    def sensitivity_map(self, init_mass=0., x_grid=None, y_grid=None, 
                        use_jax_vectorize=False):
        # prepare the new model
        self.prepare_halo_model(init_mass)

        # create the loss corresponding this new model
        self.halo_loss = Loss(self.data, self.halo_lens_image, self.halo_param, 
                              likelihood_type='chi2')

        # define the grid on which to compute sensitivity
        if self.halo_lens_image.ImageNumerics.grid_supersampling_factor > 1:
            x_grid, y_grid = self.halo_lens_image.ImageNumerics.coordinates_evaluate # 1D arrays
        elif x_grid is None or y_grid is None:
            x_grid, y_grid = self.halo_lens_image.Grid.pixel_coordinates

        # CHECKS
        if self.verbose:
            print("halo loss at edge:", self.halo_loss([init_mass, x_grid[0, 0], y_grid[0, 0]]))
            print("halo grad loss at edge:", self.halo_loss.gradient([init_mass, x_grid[0, 0], y_grid[0, 0]]))

            print("macro loss:", self.m_loss(self.p_macro))
            # print("macro grad loss:", grad(self.m_loss)(self.p_macro))

        @jit
        def sensitivity_at_pixel(x, y):
            if self.fix_macro:
                p = [init_mass, x, y]
            else:
                p = [init_mass, x, y] + self.p_macro
            grad_loss_mass = self.halo_loss.gradient(p)
            partial_deriv_mass = grad_loss_mass[0]
            return partial_deriv_mass
        
        # evaluate the sensitivity over the coordinates grid
        start = time.time()
        if use_jax_vectorize:
            sensitivity_map = jnp.vectorize(sensitivity_at_pixel)(x_grid, y_grid).block_until_ready()
        else:
            sensitivity_map = np.vectorize(sensitivity_at_pixel)(x_grid, y_grid)
        runtime = time.time() - start

        # convert to numpy array and reshape
        sensitivity_map = np.array(sensitivity_map)
        if len(sensitivity_map.shape) == 1:
            sensitivity_map = util.array2image(sensitivity_map)

        # get the coordinates where the sensitivity is the highest
        peak_indices_2d = feature.peak_local_max(-np.clip(sensitivity_map, a_min=None, a_max=0))
        
        x_coords, y_coords = x_grid[0, :], y_grid[:, 0]
        x_minima = x_coords[peak_indices_2d[:, 1]]
        y_minima = y_coords[peak_indices_2d[:, 0]]
        z_minima = sensitivity_map[peak_indices_2d[:, 0], peak_indices_2d[:, 1]]

        # if with_covariance:
        #     @jit
        #     def covariance_at_pixel(x, y):
        #         if self.fix_macro:
        #             p = [init_mass, x, y]
        #         else:
        #             p = [init_mass, x, y] + self.p_macro
        #         grad_loss_mass = self.halo_loss.hessian(p)
        #         partial_deriv_mass = grad_loss_mass[0]
        #         return partial_deriv_mass
        #     covariance_cube = np.vectorize(covariance_at_pixel)(x_grid, y_grid)

        return sensitivity_map, (x_minima, y_minima, z_minima), runtime

    def sensitivity_map_optim(self, init_mass=0., x_grid=None, y_grid=None, minimize_method='trust-krylov', **kwargs_optimizer):
        # prepare the new model
        self.prepare_halo_model(init_mass)

        # create the loss corresponding this new model
        self.halo_loss = Loss(self.data, self.halo_lens_image, self.halo_param, likelihood_type='chi2')

        # create the optimizer
        optimizer = Optimizer(self.halo_loss, self.halo_param)

        # define the grid on which to compute sensitivity
        if self.halo_lens_image.ImageNumerics.grid_supersampling_factor > 1:
            x_grid, y_grid = self.halo_lens_image.ImageNumerics.coordinates_evaluate # 1D arrays
        elif x_grid is None or y_grid is None:
            x_grid, y_grid = self.halo_lens_image.Grid.pixel_coordinates

        # CHECKS
        if self.verbose:
            print("halo loss at edge:", self.halo_loss([init_mass, x_grid[0, 0], y_grid[0, 0]]))
            print("halo grad at edge:", self.halo_loss.gradient([init_mass, x_grid[0, 0], y_grid[0, 0]]))

            print("macro loss:", self.m_loss(self.p_macro))
            # print("macro grad loss:", grad(self.m_loss)(self.p_macro))

        def sensitivity_at_pixel(x, y):
            if self.fix_macro:
                p = [init_mass, x, y]
            else:
                p = [init_mass, x, y] + self.p_macro
            best_fit, logL, _, runtime = optimizer.minimize(method=minimize_method, init_params=p, **kwargs_optimizer)
            return float(best_fit[0]), logL, runtime
        
        # evaluate the sensitivity over the coordinates grid
        start = time.time()
        sensitivity_map, logL_map, runtimes = np.vectorize(sensitivity_at_pixel)(x_grid, y_grid)
        runtime = np.sum(runtimes)

        # convert to numpy array and reshape
        sensitivity_map = np.array(sensitivity_map)
        logL_map = np.array(logL_map)
        if len(sensitivity_map.shape) == 1:
            sensitivity_map = util.array2image(sensitivity_map)
            logL_map = util.array2image(logL_map)

        # get the coordinates where the sensitivity is the highest
        peak_indices_2d = feature.peak_local_max(-np.clip(sensitivity_map, a_min=None, a_max=0))
        
        x_coords, y_coords = x_grid[0, :], y_grid[:, 0]
        x_minima = x_coords[peak_indices_2d[:, 1]]
        y_minima = y_coords[peak_indices_2d[:, 0]]
        z_minima = sensitivity_map[peak_indices_2d[:, 0], peak_indices_2d[:, 1]]

        # if with_covariance:
        #     @jit
        #     def covariance_at_pixel(x, y):
        #         if self.fix_macro:
        #             p = [init_mass, x, y]
        #         else:
        #             p = [init_mass, x, y] + self.p_macro
        #         grad_loss_mass = self.halo_loss.hessian(p)
        #         partial_deriv_mass = grad_loss_mass[0]
        #         return partial_deriv_mass
        #     covariance_cube = np.vectorize(covariance_at_pixel)(x_grid, y_grid)

        return sensitivity_map, (x_minima, y_minima, z_minima), logL_map, runtime

    def prepare_halo_model(self, init_mass):
        # value at which the gradient will be evaluated
        init_mass = init_mass

        if self.halo_profile == 'POINT_MASS':
            kwargs_halo_fixed = {}
            kwargs_halo_init = {'theta_E': init_mass, 'center_x': 0., 'center_y': 0.}
        elif self.halo_profile == 'PIXELATED_DIRAC':
            kwargs_halo_fixed = {}
            kwargs_halo_init = {'psi': init_mass, 'center_x': 0., 'center_y': 0.}
        elif self.halo_profile == 'SIS':
            kwargs_halo_fixed = {}
            kwargs_halo_init = {'theta_E': init_mass, 'center_x': 0., 'center_y': 0.}
        else:
            raise NotImplementedError(f"Halo profile must be in {self._MODELS_SUPPORTED}.")
        
        halo_mass_model_list = [self.halo_profile] + self.m_lens_image.MassModel.profile_type_list
        halo_mass_model = MassModel(halo_mass_model_list)

        grid = copy.deepcopy(self.m_lens_image.Grid)
        #grid.remove_model_grid('lens')
        psf = copy.deepcopy(self.m_lens_image.PSF)
        noise = copy.deepcopy(self.m_lens_image.Noise)
        self.halo_lens_image = LensImage(grid, psf, noise_class=noise,
                                    lens_mass_model_class=halo_mass_model,
                                    source_model_class=self.m_lens_image.SourceModel,
                                    lens_light_model_class=self.m_lens_image.LensLightModel,
                                    kwargs_numerics=self.kwargs_numerics)

        kwargs_macro = self.m_param.current_values(as_kwargs=True)
        self.p_macro = copy.deepcopy(self.m_param.current_values(as_kwargs=False)).tolist()

        if self.fix_macro:
            kwargs_fixed = {
                'kwargs_lens': [kwargs_halo_fixed] + kwargs_macro['kwargs_lens'],
                'kwargs_source': kwargs_macro['kwargs_source'],
                'kwargs_lens_light': kwargs_macro['kwargs_lens_light'],
            }
        else:
            kwargs_fixed = {
                'kwargs_lens': [kwargs_halo_fixed] + [{} for _ in range(len(kwargs_macro['kwargs_lens']))],
                'kwargs_source': [{} for _ in range(len(kwargs_macro['kwargs_source']))],
                'kwargs_lens_light': kwargs_macro['kwargs_lens_light'],
            }
            # some parameters still need to be fixed
            for i, kwargs_profile in enumerate(kwargs_macro['kwargs_lens']):
                for key, value in kwargs_profile.items():
                    if key in ['ra_0', 'dec_0']:  # external shear origin
                        kwargs_fixed['kwargs_lens'][i+1][key] = value
                    elif key in ['center_x', 'center_y']:  # macro lens center (ATTENTION if this is not wanted)
                        kwargs_fixed['kwargs_lens'][i+1][key] = value

        kwargs_init = {
            'kwargs_lens': [kwargs_halo_init] + kwargs_macro['kwargs_lens'],
            'kwargs_source': kwargs_macro['kwargs_source'],
            'kwargs_lens_light': kwargs_macro['kwargs_lens_light'],
        }
        self.halo_param = Parameters(self.halo_lens_image, kwargs_init, kwargs_fixed)

        if self.verbose:
            print("parameters:", self.halo_param.names)
            print("num. params:", self.halo_param.num_parameters)
            print("init. values:", self.halo_param.initial_values())



# Defines a general fully differentiable scalar function
# 
# Copyright (c) 2022, herculens developers and contributors

__author__ = 'aymgal'


from functools import partial
from jax import jit, grad, jacfwd, jacrev, jvp, value_and_grad


__all__ = ['Differentiable']


class Differentiable(object):

    """Abstract class that defines a function with its derivatives, typically the loss function.
    """

    def _func(self, args):
        raise NotImplementedError

    @partial(jit, static_argnums=(0,))
    def __call__(self, args):
        """alias differentiable function"""
        return self._func(args)

    @partial(jit, static_argnums=(0,))
    def function(self, args):
        return self._func(args)

    @partial(jit, static_argnums=(0,))
    def gradient(self, args):
        """gradient (first derivative) of the loss function"""
        return grad(self._func)(args)

    @partial(jit, static_argnums=(0,))
    def value_and_gradient(self, args):
        return value_and_grad(self._func)(args)

    @partial(jit, static_argnums=(0,))
    def hessian(self, args):
        """hessian (second derivative) of the loss function"""
        return jacfwd(jacrev(self._func))(args)

    @partial(jit, static_argnums=(0,))
    def hessian_vec_prod(self, args, vec):
        """hessian-vector product"""
        # forward-over-reverse (https://jax.readthedocs.io/en/latest/notebooks/autodiff_cookbook.html#hessian-vector-products-using-both-forward-and-reverse-mode)
        return jvp(grad(self._func), (args,), (vec,))[1]



# Defines the full loss function, from likelihood, prior and regularization terms
# 
# Copyright (c) 2021, herculens developers and contributors

__author__ = 'aymgal', 'austinpeel'


import numpy as np
import jax.numpy as jnp
from jax import jit
import warnings

from herculens.Inference.base_differentiable import Differentiable


__all__ = ['Loss']


class Loss(Differentiable):

    def __init__(self, prob_model, constrained_space=False, cap_value=None):
        """
        :param prob_model: probabilistic model (e.g. from numpyro) that has a
        log_prob() method that returns the full log-probability of the model
        :param constrained_space: whether or not to consider that the parameters
        (input values of log_prob()) are assumed to be in constrained or 
        unconstrained space
        """
        self._prob_model = prob_model
        self._constrained = constrained_space
        self._cap_value = cap_value

    def _func(self, args):
        """negative log-probability"""
        loss = - self._prob_model.log_prob(args, constrained=self._constrained)
        loss = jnp.nan_to_num(loss, nan=1e15, posinf=1e15, neginf=1e15)
        if self._cap_value is not None:
            loss = jnp.clip(loss, a_min=self._cap_value)
        return loss



# Defines a pixelated profile
# 
# Copyright (c) 2021, herculens developers and contributors

__author__ = 'aymgal'

import jax.numpy as jnp


__all__= ['Shapelets']


class Shapelets(object):
    """Surface brightness with a set of basis functions"""

    param_names = ['beta', 'center_x', 'center_y', 'amps']
    lower_limit_default = {'beta': 0.,  'center_x': 100., 'center_y': -100., 'amps': -1e10}
    upper_limit_default = {'beta': 1e5, 'center_x': 100., 'center_y':  100., 'amps':  1e10}
    fixed_default = {key: False for key in param_names}

    def __init__(self, n_max, function_type='gaussian'):
        if function_type == 'gaussian':
            from gigalens.jax.profiles.light.shapelets import Shapelets as ShapeletsGigaLens
            self._n_max = n_max
            self._backend = ShapeletsGigaLens(self._n_max, use_lstsq=False, interpolate=False)
        else:
            raise NotImplementedError(f"Basis function type '{function_type}' is not supported")
        self._func_type = function_type

    @property
    def maximum_order(self):
        return self._n_max

    @property
    def num_amplitudes(self):
        if self._func_type == 'gaussian':
            return int((self._n_max+1) * (self._n_max+2) / 2)

    def function(self, x, y, beta, center_x, center_y, amps):
        # convert amplitudes to GigaLens syntax
        decimal_places = len(str(self._backend.depth))
        amps_dict = {f'amp{str(i).zfill(decimal_places)}': jnp.array([amps[i]]) for i in range(self.num_amplitudes)}
        return self._backend.light(x, y, center_x, center_y, beta, **amps_dict)



# Set of utilities that extends some of numpyro functionalities
# In the future, these may be incorporated within numpyro 


import jax.numpy as jnp
from functools import partial

import numpyro
from numpyro import handlers
from numpyro.distributions import transforms, constraints
from numpyro.distributions.util import sum_rightmost
from numpyro.infer import util


def unconstrain_reparam(params, site):
    """added support for numpyro.param sites"""
    name = site["name"]
    if name in params:
        p = params[name]
        
        if site["type"] == "param":
            constraint = site["kwargs"].get("constraint", constraints.real)
            with util.helpful_support_errors(site):
                transform = transforms.biject_to(constraint)
            
            len_event_shape = len(site["kwargs"]["event_dim"])

        else:
            support = site["fn"].support
            with util.helpful_support_errors(site):
                transform = transforms.biject_to(support)
            
            # in scan, we might only want to substitute an item at index i, rather than the whole sequence
            i = site["infer"].get("_scan_current_index", None)
            if i is not None:
                event_dim_shift = transform.codomain.event_dim - t.domain.event_dim
                expected_unconstrained_dim = len(site["fn"].shape()) - event_dim_shift
                # check if p has additional time dimension
                if jnp.ndim(p) > expected_unconstrained_dim:
                    p = p[i]

            if support is constraints.real or (
                isinstance(support, constraints.independent)
                and support.base_constraint is constraints.real
            ):
                return p

            len_event_shape = len(site["fn"].event_shape)

        value = transform(p)

        # NOTE: we add the determinant term only for sampled sites
        # and only transformed parameter site values above 
        if site["type"] == "sample":
            log_det = transform.log_abs_det_jacobian(p, value)
            log_det = sum_rightmost(
                log_det, jnp.ndim(log_det) - jnp.ndim(value) + len_event_shape
            )
            numpyro.factor("_{}_log_det".format(name), log_det)

        return value

def potential_energy(model, model_args, model_kwargs, params):
    """
    This is essentially the same as numpyro.infer.util.potential_energy, but with
    the added support for numpyro.param sites, through a modification 
    of the unconstrain_reparam function.

    (EXPERIMENTAL INTERFACE) Computes potential energy of a model given unconstrained params.
    Under the hood, we will transform these unconstrained parameters to the values
    belong to the supports of the corresponding priors in `model`.

    :param model: a callable containing NumPyro primitives.
    :param tuple model_args: args provided to the model.
    :param dict model_kwargs: kwargs provided to the model.
    :param dict params: unconstrained parameters of `model`.
    :return: potential energy given unconstrained parameters.
    """
    substituted_model = handlers.substitute(
        model, substitute_fn=partial(unconstrain_reparam, params)
    )
    # no param is needed for log_density computation because we already substitute
    log_joint, model_trace = util.log_density(
        substituted_model, model_args, model_kwargs, {}
    )
    return -log_joint



# Defines a point mass
# 
# Copyright (c) 2021, herculens developers and contributors

__author__ = 'aymgal'


import numpy as np
import jax.numpy as jnp


__all__ = ['PointMass']


class PointMass(object):
    """
    class to compute the physical deflection angle of a point mass, given as an Einstein radius
    """
    param_names = ['theta_E', 'center_x', 'center_y']
    lower_limit_default = {'theta_E': 0, 'center_x': -100, 'center_y': -100}
    upper_limit_default = {'theta_E': 100, 'center_x': 100, 'center_y': 100}
    fixed_default = {key: False for key in param_names}

    def __init__(self):
        self.r_min = 10e-25
        super(PointMass, self).__init__()

    def function(self, x, y, theta_E, center_x=0, center_y=0):
        """

        :param x: x-coord (in angles)
        :param y: y-coord (in angles)
        :param theta_E: Einstein radius (in angles)
        :return: lensing potential
        """
        x_ = x - center_x
        y_ = y - center_y
        r_ = jnp.sqrt(x_**2 + y_**2)
        r  = jnp.clip(r_, a_min=self.r_min)
        phi = theta_E**2*np.log(r)
        return phi

    def derivatives(self, x, y, theta_E, center_x=0, center_y=0):
        """

        :param x: x-coord (in angles)
        :param y: y-coord (in angles)
        :param theta_E: Einstein radius (in angles)
        :return: deflection angle (in radian)
        """
        x_ = x - center_x
        y_ = y - center_y
        r_ = jnp.sqrt(x_**2 + y_**2)
        r  = jnp.clip(r_, a_min=self.r_min)
        alpha = theta_E**2/r
        return alpha*x_/r, alpha*y_/r

    def hessian(self, x, y, theta_E, center_x=0, center_y=0):
        """

        :param x: x-coord (in angles)
        :param y: y-coord (in angles)
        :param theta_E: Einstein radius (in angles)
        :return: hessian matrix (in radian)
        """
        x_ = x - center_x
        y_ = y - center_y
        C = theta_E**2
        r2_ = x_**2 + y_**2
        r2  = jnp.clip(r2_, a_min=self.r_min**2)
        f_xx = C * (y_**2-x_**2)/r2**2
        f_yy = C * (x_**2-y_**2)/r2**2
        f_xy = -C * 2*x_*y_/r2**2
        return f_xx, f_yy, f_xy



# NOTE: this file is useful for backward-compatibility only, 
# as the preferred way is now to pass the profile class directly 
# to the MassModel() constructor.

from herculens.MassModel.Profiles.gaussian_potential import Gaussian
from herculens.MassModel.Profiles.point_mass import PointMass
from herculens.MassModel.Profiles.multipole import Multipole
from herculens.MassModel.Profiles.shear import Shear, ShearGammaPsi
from herculens.MassModel.Profiles.sis import SIS
from herculens.MassModel.Profiles.sie import SIE
from herculens.MassModel.Profiles.nie import NIE
from herculens.MassModel.Profiles.epl import EPL
from herculens.MassModel.Profiles.piemd import PIEMD
from herculens.MassModel.Profiles.dpie import DPIE
from herculens.MassModel.Profiles.pixelated import (
    PixelatedPotential,
    PixelatedFixed,
    PixelatedPotentialDirac,
)

# mapping between the string name to the mass profile class.
STRING_MAPPING = {
    'EPL': EPL,
    'NIE': NIE,
    'SIE': SIE,
    'SIS': SIS,
    'DPIE': DPIE,
    'PIEMD': PIEMD,
    'GAUSSIAN': Gaussian,
    'POINT_MASS': PointMass,
    'SHEAR': Shear,
    'SHEAR_GAMMA_PSI': ShearGammaPsi,
    'MULTIPOLE': Multipole,
    'PIXELATED': PixelatedPotential,
    'PIXELATED_DIRAC': PixelatedPotentialDirac,
    'PIXELATED_FIXED': PixelatedFixed,
}

SUPPORTED_MODELS = list(STRING_MAPPING.keys())



# This file is part of Herculens, a python code for lens
# modelling and it is based on the VKL C++ code described in
# Vernardos & Koopmans 2022, which can be found here:
# https://github.com/gvernard/verykool
#
# Copyright (C) 2022 Georgios Vernardos <gvernard@astro.rug.nl>
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

__author__ = 'gvernard', 'aymgal'


import math
import numpy as np
from scipy import sparse


def vkl_operator(data_x,data_y,dpsi_xmin,dpsi_dx,dpsi_Nx,dpsi_x,dpsi_ymax,dpsi_dy,dpsi_Ny,dpsi_y,source0_dx,source0_dy):
    # Arguments:
    #  data_x: An array containing the X coordinates of the data pixels - must be the length of the data
    #  data_y: An array containing the Y coordinates of the data pixels - must be the length of the data
    #  dpsi_xmin: The minimum X coordinate of the dpsi grid
    #  dpsi_dx: The separation between two dpsi grid pixels in the X direction
    #  dpsi_Nx: The number of dpsi pixels in the X direction
    #  dpsi_x: An array containing the X coordinates of the dpsi pixels - must be the length of the data
    #  dpsi_ymax: The maximum Y coordinate of the dpsi grid
    #  dpsi_dy: The separation between two dpsi grid pixels in the Y direction
    #  dpsi_Ny: The number of dpsi pixels in the Y direction
    #  source0_x: An array of the source derivatives in the X direction at the location of the deflected data pixels (size=number of data pixels)
    #  source0_y: An array of the source derivatives in the Y direction at the location of the deflected data pixels (size=number of data pixels)
    crosses = createCrosses(data_x,data_y,dpsi_xmin,dpsi_dx,dpsi_Nx,dpsi_x,dpsi_ymax,dpsi_dy,dpsi_Ny,dpsi_y)
    operator = constructDsDpsi(crosses,source0_dx,source0_dy,dpsi_Nx,dpsi_Ny)
    return operator


def derivativeDirection(q,qmax,den):
    # Description:
    #  Helper function (called only from within this code) that calculates the finite difference coefficients and their indices.
    #  This is done for the 2nd order accuracy 1st derivatives, for both X and Y derivatives.
    # Arguments:
    #  q: the index (in X or Y direction) of the given dpsi grid pixel where the derivative needs to be calculated via finite differences
    #  qmax: the maximum index (in X or Y direction) of the dpsi grid
    #  den: the denominator to use in the finite difference coefficients, i.e. either dx or dy of the dpsi grid
    # Returns:
    #  rel_ind: the relative indices of the finite difference coefficients with respect to q
    #  coeff: the finite difference coefficients

    if q == 0:
        # FORWARD finite difference, 1st derivative, 2nd order
        rel_ind = [0,1,2]
        coeff = [-3.0/2.0, 2.0, -1.0/2.0]/den
    elif q == qmax:
        # BACKWARD finite difference, 1st derivative, 2nd order
        rel_ind = [0,-1,-2]
        coeff = [3.0/2.0, -2.0, 1.0/2.0]/den
    else:
        # CENTRAL finite difference, 1st derivative, 2nd order
        rel_ind = [-1,0,1]
        coeff = [-1.0/2.0, 0.0, 1.0/2.0]/den
    return rel_ind,coeff


def createCrosses(data_x,data_y,dpsi_xmin,dpsi_dx,dpsi_Nx,dpsi_x,dpsi_ymax,dpsi_dy,dpsi_Ny,dpsi_y):
    # Description:
    #  This function creates a 'cross' structure for each data pixel.
    # Arguments:
    #  data_x: An array containing the X coordinates of the data pixels - must be the length of the data
    #  data_y: An array containing the Y coordinates of the data pixels - must be the length of the data
    #  dpsi_xmin: The minimum X coordinate of the dpsi grid
    #  dpsi_dx: The separation between two dpsi grid pixels in the X direction
    #  dpsi_Nx: The number of dpsi pixels in the X direction
    #  dpsi_x: An array containing the X coordinates of the dpsi pixels - must be the length of the data
    #  dpsi_ymax: The maximum Y coordinate of the dpsi grid
    #  dpsi_dy: The separation between two dpsi grid pixels in the Y direction
    #  dpsi_Ny: The number of dpsi pixels in the Y direction
    # Returns:
    #  crosses: An array of 'cross' structures

    # Loop over the data pixels and assign a 'cross' structure to each.
    crosses = []
    for h in range(0,len(data_x)):
        # Step 1: find where the data pixel (ray) lies within the dpsi grid.
        #         This ray will lie between 4 pixels in the dpsi grid.
        #         Find the indices (row,col) of the top left dpsi pixel of these four.
        j0 = math.floor( (data_x[h] - dpsi_xmin)/dpsi_dx )
        i0 = math.floor( (dpsi_ymax - data_y[h])/dpsi_dy ) # y-axis is reflected (indices i start from the top)
        if j0 >= (dpsi_Nx - 1):
            j0 = dpsi_Nx - 2
        if i0 >= (dpsi_Ny - 1):
            i0 = dpsi_Ny - 2
        four_i = [i0,i0,i0+1,i0+1]
        four_j = [j0,j0+1,j0,j0+1]

        
        # Step 2: get the interpolation weights for the ray (data) with respect to the four dpsi grid pixels that contain it
        ya = data_y[h]               - dpsi_y[(i0+1)*dpsi_Nx]
        yb = dpsi_y[i0*dpsi_Nx]      - data_y[h]
        xa = data_x[h]               - dpsi_x[i0*dpsi_Nx+j0]
        xb = dpsi_x[i0*dpsi_Nx+j0+1] - data_x[h]
        den = dpsi_dx*dpsi_dy
        four_w = [xb*ya,xa*ya,xb*yb,xa*yb]/den

        # Step 3: loop over the dpsi pixel vertices and store the X and Y coefficients in the cross.
        #         The way the coefficients are ordered is defined ad-hoc, left-to-right and top-to-bottom (see notes).
        cross = {
            'i0': i0,
            'j0': j0,
            'coeff_x': np.zeros(8),
            'coeff_y': np.zeros(8)
        }
        for k in range(0,len(four_i)):
            # X derivatives
            rel_ind,coeffs = derivativeDirection(four_j[k],dpsi_Nx-1,dpsi_dx)
            for qq in range(0,len(rel_ind)):
                cross_index = (four_i[k]-i0)*4 + (four_j[k]-j0) + rel_ind[qq] + 1
                cross['coeff_x'][cross_index] += coeffs[qq]*four_w[k]

            # Y derivatives
            rel_ind,coeffs = derivativeDirection(four_i[k],dpsi_Ny-1,dpsi_dy)
            for qq in range(0,len(rel_ind)):
                cross_index = (four_i[k]-i0+rel_ind[qq]+1)*2 + (four_j[k]-j0)
                cross['coeff_y'][cross_index] += coeffs[qq]*four_w[k]
        crosses.append(cross)

    return crosses


def constructDsDpsi(crosses,source0_x,source0_y,dpsi_Nx,dpsi_Ny):
    # Description:
    #  Finds the non-zero coefficients for the term given by equation 7 in Vernardos & Koopmans 2022 for each data pixel.
    # Arguments:
    #  crosses: An array of 'cross' structures, the output of 'createCrosses' (size=number of data pixels)
    #  source0_x: An array of the source derivatives in the X direction at the location of the deflected data pixels (size=number of data pixels)
    #  source0_y: An array of the source derivatives in the Y direction at the location of the deflected data pixels (size=number of data pixels)
    # Returns:
    #  sparse_mat: An array of tuples. Each tuple contains the row index, column index, and value of the non-zero elements of the sparse matrix
    rel_i = [-1, -1,  0, 0, 0, 0,  1, 1, 1, 1, 2, 2]
    rel_j = [ 0,  1, -1, 0, 1, 2, -1, 0, 1, 2, 0, 1]
    vals = np.zeros(12)
    
    # sparse_mat = []
    sparse_values = []
    sparse_rows, sparse_cols = [], []
    for h in range(0,len(crosses)):
        src_Dx = source0_x[h]
        src_Dy = source0_y[h]

        vals[0]  = crosses[h]['coeff_y'][0]*src_Dy
        vals[1]  = crosses[h]['coeff_y'][1]*src_Dy
        vals[2]  = crosses[h]['coeff_x'][0]*src_Dx
        vals[3]  = crosses[h]['coeff_x'][1]*src_Dx + crosses[h]['coeff_y'][2]*src_Dy
        vals[4]  = crosses[h]['coeff_x'][2]*src_Dx + crosses[h]['coeff_y'][3]*src_Dy
        vals[5]  = crosses[h]['coeff_x'][3]*src_Dx
        vals[6]  = crosses[h]['coeff_x'][4]*src_Dx
        vals[7]  = crosses[h]['coeff_x'][5]*src_Dx + crosses[h]['coeff_y'][4]*src_Dy
        vals[8]  = crosses[h]['coeff_x'][6]*src_Dx + crosses[h]['coeff_y'][5]*src_Dy
        vals[9]  = crosses[h]['coeff_x'][7]*src_Dx
        vals[10] = crosses[h]['coeff_y'][6]*src_Dy
        vals[11] = crosses[h]['coeff_y'][7]*src_Dy
        
        for q in range(0,12):
            if vals[q] != 0:
                col_index = (crosses[h]['i0']+rel_i[q])*dpsi_Nx + (crosses[h]['j0']+rel_j[q])
                
                # sparse_mat.append( (h,col_index,vals[q]) )
                sparse_values.append(vals[q])
                sparse_rows.append(h)
                sparse_cols.append(col_index)
            
            vals[q] = 0 # Need to set the values of the 12 coefficients back to zero

    # populate the sparse matrix
    Ndata = len(crosses)
    Ndpsi = dpsi_Nx*dpsi_Ny
    dense_shape = (Ndata, Ndpsi)
    DsDpsi_matrix = sparse.csr_matrix((sparse_values, (sparse_rows, sparse_cols)), 
                                      shape=dense_shape)
    return DsDpsi_matrix



# Utility functions for defining linear operators
# 
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2020, SLITronomy developers and contributors

__author__ = 'austinpeel', 'gvernard', 'aymgal'


import numpy as np
from scipy import sparse, linalg
import findiff

from utax.interpolation import BicubicInterpolator as Interpolator
from herculens.Util import util


def build_bilinear_interpol_matrix(x_grid_1d_in, y_grid_1d_in, x_grid_1d_out, 
                                   y_grid_1d_out, warning=True):
    """
    Only works with square input and output grids.
    Author: austinpeel, originally for the package `slitronomy`.
    """
    # Standardize inputs for vectorization
    x_grid_1d_out = np.atleast_1d(x_grid_1d_out)
    y_grid_1d_out = np.atleast_1d(y_grid_1d_out)
    assert len(x_grid_1d_out) == len(y_grid_1d_out), "Input arrays must be the same size."
    num_pix_out = len(x_grid_1d_out)
    
    # Compute bin edges so that (x_coord, y_coord) lie at the grid centers
    num_pix = int(np.sqrt(x_grid_1d_in.size))
    delta_pix = np.abs(x_grid_1d_in[0] - x_grid_1d_in[1])
    half_pix = delta_pix / 2.

    x_coord = x_grid_1d_in[:num_pix]
    x_dir = -1 if x_coord[0] > x_coord[-1] else 1  # Handle x-axis inversion
    x_lower = x_coord[0] - x_dir * half_pix
    x_upper = x_coord[-1] + x_dir * half_pix
    xbins = np.linspace(x_lower, x_upper, num_pix + 1)

    y_coord = y_grid_1d_in[::num_pix]
    y_dir = -1 if y_coord[0] > y_coord[-1] else 1  # Handle y-axis inversion
    y_lower = y_coord[0] - y_dir * half_pix
    y_upper = y_coord[-1] + y_dir * half_pix
    ybins = np.linspace(y_lower, y_upper, num_pix + 1)

    # Keep only coordinates that fall within the output grid
    x_min, x_max = [x_lower, x_upper][::x_dir]
    y_min, y_max = [y_lower, y_upper][::y_dir]
    selection = ((x_grid_1d_out > x_min) & (x_grid_1d_out < x_max) &
                 (y_grid_1d_out > y_min) & (y_grid_1d_out < y_max))
    if np.any(1 - selection.astype(int)):
        x_grid_1d_out = x_grid_1d_out[selection]
        y_grid_1d_out = y_grid_1d_out[selection]
        num_pix_out = len(x_grid_1d_out)

    # Find the (1D) output pixel that (x_grid_1d_out, y_grid_1d_out) falls in
    index_x = np.digitize(x_grid_1d_out, xbins) - 1
    index_y = np.digitize(y_grid_1d_out, ybins) - 1
    index_1 = index_x + index_y * num_pix

    # Compute distances between input and output grid points
    dx = x_grid_1d_out - x_grid_1d_in[index_1]
    dy = y_grid_1d_out - y_grid_1d_in[index_1]

    # Find the three other nearest pixels (may end up out of bounds)
    index_2 = index_1 + x_dir * np.sign(dx).astype(int)
    index_3 = index_1 + y_dir * np.sign(dy).astype(int) * num_pix
    index_4 = index_2 + y_dir * np.sign(dy).astype(int) * num_pix

    # Treat these index arrays as four sets stacked vertically
    # Prepare to mask out out-of-bounds pixels as well as repeats
    # The former is important for the csr_matrix to be generated correctly
    max_index = x_grid_1d_in.size - 1  # Upper index bound
    mask = np.ones((4, num_pix_out), dtype=bool)  # Mask for the coordinates

    # Mask out any neighboring pixels that end up out of bounds
    mask[1, np.where((index_2 < 0) | (index_2 > max_index))[0]] = False
    mask[2, np.where((index_3 < 0) | (index_3 > max_index))[0]] = False
    mask[3, np.where((index_4 < 0) | (index_4 > max_index))[0]] = False

    # Mask any repeated pixels (2 or 3x) arising from unlucky grid alignment
    # zero_dx = list(np.where(dx == 0)[0])
    # zero_dy = list(np.where(dy == 0)[0])
    # unique, counts = np.unique(zero_dx + zero_dy, return_counts=True)
    # repeat_row = [ii + 1 for c in counts for ii in range(0, 3, 3 - c)]
    # repeat_col = [u for (u, c) in zip(unique, counts) for _ in range(c + 1)]
    # mask[(repeat_row, repeat_col)] = False  # TODO: this leads to strange lines

    # Generate 2D indices of non-zero elements for the sparse matrix
    row = np.tile(np.nonzero(selection)[0], (4, 1))
    col = np.array([index_1, index_2, index_3, index_4])

    # Compute bilinear weights like in Treu & Koopmans (2004)
    col[~mask] = 0  # Avoid accessing values out of bounds
    dist_x = (np.tile(x_grid_1d_out, (4, 1)) - x_grid_1d_in[col]) / delta_pix
    dist_y = (np.tile(y_grid_1d_out, (4, 1)) - y_grid_1d_in[col]) / delta_pix
    weight = (1 - np.abs(dist_x)) * (1 - np.abs(dist_y))

    # Make sure the weights are properly normalized
    # This step is only necessary where the mask has excluded source pixels
    norm = np.expand_dims(np.sum(weight, axis=0, where=mask), 0)
    weight = weight / norm

    if warning:
        if np.any(weight[mask] < 0):
            num_neg = np.sum((weight[mask] < 0).astype(int))
            print("Warning : {} weights are negative.".format(num_neg))

    indices, weights = (row[mask], col[mask]), weight[mask]

    dense_shape = (x_grid_1d_out.size, x_grid_1d_in.size)
    interpol_matrix = sparse.csr_matrix((weights, indices), shape=dense_shape)
    interpol_norm = np.squeeze(np.maximum(1, interpol_matrix.sum(axis=0)).A)
    return interpol_matrix, interpol_norm


def build_DsD_matrix(smooth_lens_image, smooth_kwargs_params, hybrid_lens_image=None):
    """this functions build the full operator from Koopmans 2005"""
    # data grid
    # x_coords, y_coords = smooth_lens_image.Grid.pixel_axes
    x_grid, y_grid = smooth_lens_image.Grid.pixel_coordinates
    num_pix_x, num_pix_y = smooth_lens_image.Grid.num_pixel_axes
    pixel_width = smooth_lens_image.Grid.pixel_width

    # pixelated lens model grid (TODO: implement interpolation on potential grid)
    # x_coords_pot, y_coords_pot = hybrid_lens_image.Grid.model_pixel_axes('lens')
    # x_grid_pot, y_grid_pot = hybrid_lens_image.Grid.model_pixel_coordinates('lens')

    # numerics grid, for intermediate computation on a higher resolution grid
    x_grid_num, y_grid_num = smooth_lens_image.ImageNumerics.coordinates_evaluate
    x_grid_num = util.array2image(x_grid_num)
    y_grid_num = util.array2image(y_grid_num)
    x_coords_num, y_coords_num = x_grid_num[0, :], y_grid_num[:, 0]
    # pixel_width_num = np.abs(x_coords_num[1] - x_coords_num[0])
    
    # get the pixelated source in source plane,
    # on the highest resolution grid possible (it will use )
    smooth_source = smooth_lens_image.SourceModel.surface_brightness(
        x_grid_num, y_grid_num, smooth_kwargs_params['kwargs_source'])
    interp_source = Interpolator(y_coords_num, x_coords_num, smooth_source)

    # compute its derivatives *on source plane*
    grad_s_x_srcplane = interp_source(y_grid_num, x_grid_num, dy=1)
    grad_s_y_srcplane = interp_source(y_grid_num, x_grid_num, dx=1)
    # grad_s_srcplane = np.sqrt(grad_s_x_srcplane**2 + grad_s_y_srcplane**2)

    # setup the Interpolator to read on data pixels
    interp_grad_s_x = Interpolator(y_coords_num, x_coords_num, grad_s_x_srcplane)
    interp_grad_s_y = Interpolator(y_coords_num, x_coords_num, grad_s_y_srcplane)

    # use the lens equation to ray shoot the coordinates of the data grid
    x_src, y_src = smooth_lens_image.MassModel.ray_shooting(
        x_grid, y_grid, smooth_kwargs_params['kwargs_lens'])

    # evaluate the resulting arrays on that grid
    grad_s_x = interp_grad_s_x(y_src, x_src)
    grad_s_y = interp_grad_s_y(y_src, x_src)
    # grad_s = np.sqrt(grad_s_x**2 + grad_s_y**2)

    # proper flux units
    grad_s_x *= pixel_width**2
    grad_s_y *= pixel_width**2

    # put them into sparse diagonal matrices
    D_s_x = sparse.diags([grad_s_x.flatten()], [0])
    D_s_y = sparse.diags([grad_s_y.flatten()], [0])

    # compute the potential derivative operator as two matrices D_x, D_y
    step_size = pixel_width # step size
    order = 1 # first-order derivative
    accuracy = 2 # accuracy of the finite difference scheme (2-points, 4-points, etc.)
    d_dx_class = findiff.FinDiff(1, step_size, order, acc=accuracy)
    d_dy_class = findiff.FinDiff(0, step_size, order, acc=accuracy)
    D_x = d_dx_class.matrix((num_pix_x, num_pix_y))
    D_y = d_dy_class.matrix((num_pix_x, num_pix_y))  # sparse matrices
    
    # join the source and potential derivatives operators
    # through minus their 'scalar' product (Eq. A6 from Koopmans 2005)
    DsD = - D_s_x.dot(D_x) - D_s_y.dot(D_y)

    # we also return the gradient of the source after being ray-traced to the data grid
    return DsD, grad_s_x, grad_s_y




# Handles different method to optimize a loss function
# 
# Copyright (c) 2021, herculens developers and contributors

__author__ = 'aymgal', 'austinpeel'


import time
import warnings
import numpy as np
import jax
import optax
from scipy import optimize
from scipy.optimize import Bounds
from tqdm import tqdm
from copy import deepcopy

from herculens.Inference.legacy.base_inference import Inference

__all__ = ['Optimizer']


class Optimizer(Inference):
    """Class that handles optimization tasks, i.e. finding best-fit point estimates of parameters
    It currently handles:
    - a subset of scipy.optimize.minimize routines, using first and second order derivatives when required
    - a particle swarm optimizer (PSO), implemented in lenstronomy
    """

    _supported_scipy_methods = ['Nelder-Mead', 'BFGS', 'Newton-CG', 'trust-krylov', 'trust-exact', 'trust-constr']

    # @property
    # def loss_history(self):
    #     if not hasattr(self, '_metrics'):
    #         raise ValueError("You must run the optimizer at least once to access the history")
    #     return self._metrics.loss_history

    # @property
    # def param_history(self):
    #     if not hasattr(self, '_metrics'):
    #         raise ValueError("You must run the optimizer at least once to access the history")
    #     return self._metrics.param_history

    def minimize(self, method='BFGS', maxiter=None, init_params=None,
                 restart_from_init=False, use_exact_hessian_if_allowed=False,
                 multi_start_from_prior=False, num_multi_start=1, seed_multi_start=None,
                 progress_bar=True, return_param_history=False):
        # TODO: should we call once / a few times all jitted functions before optimization, to potentially speed things up?
        metrics = MinimizeMetrics(self._loss, method, with_param_history=return_param_history)
        if multi_start_from_prior is False:
            num_multi_start = 1
        if num_multi_start == 1:
            if init_params is None:
                init_params = self._param.current_values(as_kwargs=False, restart=restart_from_init, copy=True)
            init_samples = np.asarray(init_params)[None, :]
        else:
            init_samples = self._param.draw_prior_samples(num_samples=num_multi_start, 
                                                          seed=seed_multi_start)
        exact_hessian = use_exact_hessian_if_allowed

        start = time.time()
        best_fit_list = []
        logL_best_fit_list = []
        loss_history_list = []
        param_history_list = []
        extra_fields_list = []
        for n in self._for_loop(range(num_multi_start), progress_bar, 
                                total=num_multi_start, 
                                desc=f"minimize.{method}"):
            init_params_n = init_samples[n, :]
            best_fit_n, extra_fields_n = self._run_scipy_minimizer(init_params_n, 
                                                                   method, maxiter, metrics,
                                                                   exact_hessian)
            if metrics.loss_history == []:
                warnings.warn("The loss history does not contain any value")
            best_fit_list.append(best_fit_n)
            logL_best_fit_list.append(self.log_probability(best_fit_n))
            loss_history_list.append(metrics.get_loss_history())
            param_history_list.append(metrics.get_param_history())
            extra_fields_list.append(extra_fields_n)

        # select the best fit among the multi start runs
        if num_multi_start > 1:
            index = np.argmax(logL_best_fit_list)
        else:
            index = 0
        best_fit = best_fit_list[index]
        logL_best_fit = logL_best_fit_list[index]
        extra_fields  = extra_fields_list[index]

        runtime = time.time() - start

        extra_fields['best_fit_index'] = index
        extra_fields['loss_history'] = loss_history_list[index]
        extra_fields['loss_history_list'] = loss_history_list
        if return_param_history is True:
            extra_fields['param_history'] = param_history_list[index]
            extra_fields['param_history_list'] = param_history_list  # maybe too memory consuming?
        self._param.set_best_fit(best_fit)
        return best_fit, logL_best_fit, extra_fields, runtime

    def _run_scipy_minimizer(self, x0, method, maxiter, callback, exact_hessian):
        callback.reset()
        if method not in self._supported_scipy_methods:
            raise ValueError(f"Minimize method '{method}' is not supported.")
        # here we only put select the kwargs related to the chosen method 
        extra_kwargs = {}
        if method in ['BFGS']:
            extra_kwargs['jac'] = self._loss.gradient
        elif method in ['Newton-CG', 'trust-krylov', 'trust-exact', 'trust-constr']:
            extra_kwargs['jac'] = self._loss.gradient
            if method == 'trust-exact' or exact_hessian is True:
                extra_kwargs['hess'] = self._loss.hessian
            else:
                extra_kwargs['hessp'] = self._loss.hessian_vec_prod
            if method == 'trust-constr':
                extra_kwargs['bounds'] = Bounds(*self._param.bounds)
        if maxiter is not None:
            extra_kwargs['options'] = {'maxiter': maxiter}
        res = optimize.minimize(self._loss, x0, method=method, callback=callback, 
                                **extra_kwargs)
        extra_fields = {'result_class': res, 'jac': None, 'hess': None, 'hess_inv': None}
        for key in extra_fields:
            if hasattr(res, key):
                extra_fields[key] = getattr(res, key)
        return res.x, extra_fields

    def optax(self, algorithm='adabelief', max_iterations=100, min_iterations=None,
              init_learning_rate=1e-2, schedule_learning_rate=True, 
              restart_from_init=False, stop_at_loss_increase=False, 
              progress_bar=True, return_param_history=False):
        if min_iterations is None:
            min_iterations = max_iterations
        if schedule_learning_rate is True:
            # Exponential decay of the learning rate
            scheduler = optax.exponential_decay(
                init_value=init_learning_rate, 
                decay_rate=0.99, # TODO: this has never been fine-tuned (taken from optax examples)
                transition_steps=max_iterations)

            if algorithm.lower() == 'adabelief':
                scale_algo = optax.scale_by_belief()
            elif algorithm.lower() == 'radam':
                scale_algo = optax.scale_by_radam()
            elif algorithm.lower() == 'adam':
                scale_algo = optax.scale_by_adam()
            else:
                raise ValueError(f"Optax algorithm '{algorithm}' is not supported")

            # Combining gradient transforms using `optax.chain`
            optim = optax.chain(
                #optax.clip_by_global_norm(1.0),  # clip by the gradient by the global norm # TODO: what is this used for?
                scale_algo,  # use the updates from the chosen optimizer
                optax.scale_by_schedule(scheduler),  # Use the learning rate from the scheduler
                optax.scale(-1.)  # because gradient *descent*
            )
        else:
            if algorithm.lower() == 'adabelief':
                optim = optax.adabelief(init_learning_rate)
            elif algorithm.lower() == 'radam':
                optim = optax.radam(init_learning_rate)
            elif algorithm.lower() == 'adam':
                optim = optax.adam(init_learning_rate)
            else:
                raise ValueError(f"Optax algorithm '{algorithm}' is not supported")

        # Initialise optimizer state
        params = self._param.current_values(as_kwargs=False, restart=restart_from_init, copy=True)
        opt_state = optim.init(params)
        prev_params, prev_loss_val = params, 1e10

        @jax.jit
        def gradient_step(params, opt_state):
            #loss_val, grads = jax.value_and_grad(self._loss)(params)
            loss_val, grads = self._loss.value_and_gradient(params)
            updates, opt_state = optim.update(grads, opt_state, params)
            params = optax.apply_updates(params, updates)
            return params, opt_state, loss_val

        # Gradient descent loop
        param_history = []
        loss_history = []
        start_time = time.time()
        for i in self._for_loop(range(max_iterations), progress_bar, 
                                total=max_iterations, 
                                desc=f"optax.{algorithm}"):
            params, opt_state, loss_val = gradient_step(params, opt_state)
            if stop_at_loss_increase and i > min_iterations and loss > prev_loss:
                params, loss_val = prev_params, prev_loss_val
                break
            else:
                loss_history.append(loss_val)
                prev_params, prev_loss_val = params, loss_val
            if return_param_history is True:
                param_history.append(params)
        runtime = time.time() - start_time
        best_fit = params
        logL_best_fit = self.log_probability(best_fit)
        extra_fields = {'loss_history': np.array(loss_history)}  # TODO: use optax.second_order module to compute diagonal of Hessian?
        if return_param_history is True:
            extra_fields['param_history'] = param_history
        self._param.set_best_fit(best_fit)
        return best_fit, logL_best_fit, extra_fields, runtime

    def gradient_descent(self, num_iterations=100, step_size=1e-2, restart_from_init=False):
        import jax.numpy as jnp
        # Initialise optimizer state
        params = self._param.current_values(as_kwargs=False, restart=restart_from_init, copy=True)
        
        # Gradient descent loop
        start_time = time.time()
        for _ in range(num_iterations):
            params -= step_size * self._loss.gradient(params)
        runtime = time.time() - start_time
        best_fit = params
        logL_best_fit = self.log_probability(best_fit)
        extra_fields = {}  # TODO: use optax.second_order module to compute diagonal of Hessian?
        self._param.set_best_fit(best_fit)
        return best_fit, logL_best_fit, extra_fields, runtime

    def pso(self, n_particles=100, n_iterations=100, restart_from_init=False, n_threads=1):
        """particle swarm optimization, implemented in lenstronomy"""
        from lenstronomy.Sampling.Samplers.pso import ParticleSwarmOptimizer
        from lenstronomy.Sampling.Pool.pool import choose_pool
        
        pool = choose_pool(mpi=False, processes=n_threads, use_dill=True)
        lowers, uppers = self._param.bounds
        if np.any(np.isnan(lowers)):
            raise ValueError("PSO needs lower and upper bounds, i.e. prior distributions with a finite support")
        optimizer = ParticleSwarmOptimizer(self.log_likelihood, 
                                           lowers, uppers, n_particles, pool=pool)
        init_params = self._param.current_values(as_kwargs=False, restart=restart_from_init)
        optimizer.set_global_best(init_params, [0]*len(init_params), - self._loss(init_params))
        start = time.time()
        best_fit, [chi2_list, pos_list, vel_list] = optimizer.optimize(n_iterations)
        runtime = time.time() - start
        logL_best_fit = float(chi2_list[-1])
        extra_fields = {'chi2_list': chi2_list, 'pos_list': pos_list, 'vel_list': vel_list}
        self._param.set_best_fit(best_fit)
        return best_fit, logL_best_fit, extra_fields, runtime


    @staticmethod
    def _for_loop(iterable, progress_bar_bool, **tqdm_kwargs):
        if progress_bar_bool is True:
            return tqdm(iterable, **tqdm_kwargs)
        else:
            return iterable


class MinimizeMetrics(object):
    """simple callable class used as callback in scipy.optimize.minimize method"""
    
    def __init__(self, func, method, with_param_history=False):
        self._func = func
        if method == 'trust-constr':
            self._call = self._call_2args
        else:
            self._call = self._call_1arg
        self._with_param_history = with_param_history
        self.reset()

    def reset(self):
        self.loss_history = []
        self.param_history = []

    def get_loss_history(self, copy=True):
        return deepcopy(self.loss_history) if copy else self.loss_history

    def get_param_history(self, copy=True):
        return deepcopy(self.param_history) if copy else self.param_history

    def __call__(self, *args, **kwargs):
        return self._call(*args, **kwargs)
        
    def _call_1arg(self, x):
        self.loss_history.append(float(self._func(x)))
        if self._with_param_history:
            self.param_history.append(x)

    def _call_2args(self, x, state):
        # Input state parameter is necessary for 'trust-constr' method
        # You can use it to stop execution early by returning True
        self.loss_history.append(float(self._func(x)))
        if self._with_param_history:
            self.param_history.append(x)



# Utility functions
# 
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2020, SLITronomy developers and contributors

__author__ = 'aymgal'


import numpy as np
import warnings
import matplotlib.pyplot as plt
from matplotlib import ticker
from mpl_toolkits.axes_grid1 import make_axes_locatable


def plot_minimize_history(parameters, opt_extra_fields, max_num_params=6):
    # tood: implement support for multi-start optimization
    fig, axes = plt.subplots(1, 2, figsize=(10, 6))
    ax = axes[0]
    ax.plot(range(len(opt_extra_fields['loss_history'])), opt_extra_fields['loss_history'])
    ax.set_ylabel("Loss")
    ax.set_xlabel("Iteration")
    ax = axes[1]
    if 'param_history' in opt_extra_fields:
        param_history = np.array(opt_extra_fields['param_history'])
        for i in range(min(len(parameters.names), max_num_params)):
            ax.plot(range(len(opt_extra_fields['loss_history'])), 
                    (param_history[:, i] - param_history[-1, i]) / param_history[-1, i], 
                    label=parameters.symbols[i])
        ax.set_ylabel("Parameter trace")
        ax.set_xlabel("Iteration")
        ax.legend(loc='upper right')
    else:
        warnings.warn("No `'param_history'` found in the extra fields (use `return_param_history=True` in Optimizer).")
        ax.axis('off')
    fig.tight_layout()
    return fig

def std_colorbar(mappable, label=None, fontsize=12, label_kwargs={}, **colorbar_kwargs):
    cb = plt.colorbar(mappable, **colorbar_kwargs)
    if label is not None:
        colorbar_kwargs.pop('label', None)
        cb.set_label(label, fontsize=fontsize, **label_kwargs)
    return cb

def std_colorbar_residuals(mappable, res_map, vmin, vmax, label=None, fontsize=12, 
                           label_kwargs={}, **colorbar_kwargs):
    if res_map.min() < vmin and res_map.max() > vmax:
        cb_extend = 'both'
    elif res_map.min() < vmin:
        cb_extend = 'min'
    elif res_map.max() > vmax:
        cb_extend = 'max'
    else:
        cb_extend = 'neither'
    colorbar_kwargs.update({'extend': cb_extend})
    return std_colorbar(mappable, label=label, fontsize=fontsize, 
                        label_kwargs=label_kwargs, **colorbar_kwargs)

def nice_colorbar(mappable, position='right', pad=0.1, size='5%', label=None, fontsize=12, 
                  invisible=False, max_nbins=None,
                  divider_kwargs={}, colorbar_kwargs={}, label_kwargs={}):
    divider_kwargs.update({'position': position, 'pad': pad, 'size': size})
    ax = mappable.axes
    divider = make_axes_locatable(ax)
    cax = divider.append_axes(**divider_kwargs)
    if invisible:
        cax.axis('off')
        return None
    cb = plt.colorbar(mappable, cax=cax, **colorbar_kwargs)
    if label is not None:
        colorbar_kwargs.pop('label', None)
        cb.set_label(label, fontsize=fontsize, **label_kwargs)
    if position == 'top':
        cax.xaxis.set_ticks_position('top')
    if max_nbins is not None:
        cb.locator = ticker.MaxNLocator(nbins=max_nbins)
        cb.update_ticks()
    return cb

def nice_colorbar_residuals(mappable, res_map, vmin, vmax, position='right', pad=0.1, size='5%', 
                            invisible=False, label=None, fontsize=12,
                            divider_kwargs={}, colorbar_kwargs={}, label_kwargs={}):
    if res_map.min() < vmin and res_map.max() > vmax:
        cb_extend = 'both'
    elif res_map.min() < vmin:
        cb_extend = 'min'
    elif res_map.max() > vmax:
        cb_extend = 'max'
    else:
        cb_extend = 'neither'
    colorbar_kwargs.update({'extend': cb_extend})
    return nice_colorbar(mappable, position=position, pad=pad, size=size, label=label, fontsize=fontsize,
                  invisible=invisible, colorbar_kwargs=colorbar_kwargs, label_kwargs=label_kwargs,
                  divider_kwargs=divider_kwargs)

def contour_with_legend(ax, array, kwargs_contour={}, kwargs_legend={}, logscale=True):
    """
    Add contours to an axis with a legend that actually works.
    Works with a single array or a list of arrays.
    """
    if (isinstance(array, np.ndarray) and array.ndim == 3) or (isinstance(array, (list, tuple)) and len(array) > 1):
        if isinstance(kwargs_contour, dict):
            kwargs_contour_list = [kwargs_contour] * array.shape[0]
        else:
            kwargs_contour_list = kwargs_contour
        if len(kwargs_contour_list) != len(array):
            raise ValueError("The number of contour kwargs must match the number of arrays.")
        array_list = array
    else:
        array_list = [array]
        kwargs_contour_list = [kwargs_contour]
    legend_handles = []
    for i, (arr, kw) in enumerate(zip(array_list, kwargs_contour_list)):
        label = kw.pop('label', f"Contour {i}")
        arr_transf = np.log10(arr) if logscale else arr
        ax.contour(
            arr_transf, **kw,
        )
        legend_handles.append(
            plt.Line2D(
                [0], [0], 
                linestyle=kw.get('linestyles', '-'),
                linewidth=kw.get('linewidths', 1),
                color=kw.get('colors', 'black'), 
                label=label,
            )
        )
    ax.legend(
        handles=legend_handles,
        **kwargs_legend,
    )

# def clip(data, nsigma):
#     """
#     Iteratively removes data until all is within nsigma of the median, then returns the median and std
#     author: Cameron Lemon
#     """
#     lennewdata = 0
#     lenolddata = data.size
#     while lenolddata>lennewdata:
#         lenolddata = data.size
#         data = data[np.where((data<np.nanmedian(data)+nsigma*np.nanstd(data))&(data>np.nanmedian(data)-nsigma*np.nanstd(data)))]
#         lennewdata = data.size
#     return np.median(data), np.std(data)
# 18 h 08
# bg, sigma = clip(data, 4.)
# lenspixels = data[data>bg+3*sigma]
# ax.imshow(data, origin='lower', cmap=cm, norm=LogNorm(vmax=np.nanpercentile(lenspixels, 95.), vmin=bg))



# Defines en elliptical power law profile
# 
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the LensModel.Profiles module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'dangilman', 'ntessore', 'austinpeel', 'aymgal'


import numpy as np
import jax.numpy as jnp
from herculens.Util import util, param_util, jax_util


__all__ = ['EPL', 'EPLMajorAxis']


class EPL(object):
    """
    Elliptical Power Law
    kappa = (2-t)/2*(b/r)^t
    where t = gamma - 1
    """
    param_names = ['gamma', 'theta_E', 'e1', 'e2', 'center_x', 'center_y']
    lower_limit_default = {'gamma': 1, 'theta_E': 0, 'e1': -0.5, 'e2': -0.5, 'center_x': -100, 'center_y': -100}
    upper_limit_default = {'gamma': 3, 'theta_E': 10, 'e1': 0.5, 'e2': 0.5, 'center_x': 100, 'center_y': 100}
    fixed_default = {key: False for key in param_names}
    
    def __init__(self, no_complex_numbers=False):
        self.epl_major_axis = EPLMajorAxis(no_complex_numbers)

    def param_conv(self, theta_E, e1, e2, gamma):
        """
        convert parameters from R = r sqrt(1 − e*cos(2*phi)) to
        R = sqrt(q^2 x^2 + y^2)

        :param theta_E: Einstein radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param t: power law slope
        :return: critical radius b, slope t, axis ratio q, orientation angle phi_G
        """

        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        theta_E_conv = self._theta_E_q_convert(theta_E, q)
        b = theta_E_conv * jnp.sqrt((1. + q**2) / 2.)
        t = gamma - 1.
        return b, t, q, phi_G

    def function(self, x, y, theta_E, e1, e2, gamma, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param t: power law slope
        :param center_x: profile center
        :param center_y: profile center
        :return: lensing potential
        """
        b, t, q, phi_G = self.param_conv(theta_E, e1, e2, gamma)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_G)
        # evaluate
        f_ = self.epl_major_axis.function(x__, y__, b, t, q)
        # rotate back
        return f_

    def derivatives(self, x, y, theta_E, e1, e2, gamma, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param t: power law slope
        :param center_x: profile center
        :param center_y: profile center
        :return: alpha_x, alpha_y
        """
        b, t, q, phi_G = self.param_conv(theta_E, e1, e2, gamma)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_G)
        # evaluate
        f__x, f__y = self.epl_major_axis.derivatives(x__, y__, b, t, q)
        # rotate back
        f_x, f_y = util.rotate(f__x, f__y, -phi_G)
        return f_x, f_y

    def hessian(self, x, y, theta_E, e1, e2, gamma, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param t: power law slope
        :param center_x: profile center
        :param center_y: profile center
        :return: f_xx, f_yy, f_xy
        """

        b, t, q, phi_G = self.param_conv(theta_E, e1, e2, gamma)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_G)
        # evaluate
        f__xx, f__yy, f__xy = self.epl_major_axis.hessian(x__, y__, b, t, q)
        # rotate back
        kappa = 1./2 * (f__xx + f__yy)
        gamma1__ = 1./2 * (f__xx - f__yy)
        gamma2__ = f__xy
        gamma1 = jnp.cos(2 * phi_G) * gamma1__ - jnp.sin(2 * phi_G) * gamma2__
        gamma2 = +jnp.sin(2 * phi_G) * gamma1__ + jnp.cos(2 * phi_G) * gamma2__
        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        return f_xx, f_yy, f_xy

    def _theta_E_q_convert(self, theta_E, q):
        """
        converts a spherical averaged Einstein radius to an elliptical (major axis) Einstein radius.
        This then follows the convention of the SPEMD profile in lenstronomy.
        (theta_E / theta_E_gravlens) = sqrt[ (1+q^2) / (2 q) ]

        :param theta_E: Einstein radius in lenstronomy conventions
        :param q: axis ratio minor/major
        :return: theta_E in convention of kappa=  b *(q2(s2 + x2) + y2􏰉)−1/2
        """
        theta_E_new = theta_E / (jnp.sqrt((1. + q**2) / (2. * q)))
        return theta_E_new


class EPLMajorAxis(object):
    """
    This class contains the function and the derivatives of the
    elliptical power law.

    kappa = (2-t)/2 * [b/sqrt(q^2 x^2 + y^2)]^t
    where t = gamma - 1 (from EPL class)
    Tessore & Metcalf (2015), https://arxiv.org/abs/1507.01819
    """
    param_names = ['b', 't', 'q', 'center_x', 'center_y']

    def __init__(self, no_complex_numbers):
        self._no_complex = no_complex_numbers

    def function(self, x, y, b, t, q):
        """
        returns the lensing potential
        """
        # deflection from method
        alpha_x, alpha_y = self.derivatives(x, y, b, t, q)

        # deflection potential, eq. (15)
        psi = (x * alpha_x + y * alpha_y) / (2. - t)

        return psi

    def derivatives(self, x, y, b, t, q):
        """
        returns the deflection
        """
        if self._no_complex:  # Using real numbers
            # elliptical radius, eq. (5) of Tessore et al. 2015
            R = jnp.hypot(q * x, y)
            # deflection, eq. (22)
            f_real, f_imag = jax_util.omega_real(x, y, t, q, nmax=20)
            prefac = (2 * b) / (1 + q) * ((b / R) ** (t - 1))
            alpha_real = prefac * f_real
            alpha_imag = prefac * f_imag

        else:  # Using complex numbers
            # elliptical radius, eq. (5) of Tessore et al. 2015
            z = q * x + y * 1j
            R = jnp.abs(z)
            # deflection, eq. (22)
            alpha = 2. / (1. + q) * (b / R)**t * jax_util.R_omega(z, t, q, nmax=20)
            # return real and imaginary part
            alpha_real = alpha.real # jnp.nan_to_num(alpha.real, posinf=1e8, neginf=-1e8)
            alpha_imag = alpha.imag # jnp.nan_to_num(alpha.imag, posinf=1e8, neginf=-1e8)

        return alpha_real, alpha_imag

    def hessian(self, x, y, b, t, q):
        """
        returns the Hessian matrix of the lensing potential
        """
        R = jnp.hypot(q * x, y)
        r = jnp.hypot(x, y)

        cos, sin = x / r, y / r
        cos2, sin2 = cos * cos * 2 - 1., sin * cos * 2

        # convergence, eq. (2)
        kappa = (2. - t) / 2. * (b / R)**t
        # kappa = jnp.nan_to_num(kappa, posinf=1e8, neginf=-1e8)

        # deflection via method
        alpha_x, alpha_y = self.derivatives(x, y, b, t, q)

        # shear, eq. (17), corrected version from arXiv/corrigendum
        gamma_1 = (1. - t) * (alpha_x * cos - alpha_y * sin) / r - kappa * cos2
        gamma_2 = (1. - t) * (alpha_y * cos + alpha_x * sin) / r - kappa * sin2
        gamma_1 = jnp.nan_to_num(gamma_1, posinf=1e8, neginf=-1e8)
        gamma_2 = jnp.nan_to_num(gamma_2, posinf=1e8, neginf=-1e8)

        # second derivatives from convergence and shear
        f_xx = kappa + gamma_1
        f_yy = kappa - gamma_1
        f_xy = gamma_2

        return f_xx, f_yy, f_xy



# Describes a mass model, as a list of mass profiles
# 
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the LensModel module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'austinpeel', 'aymgal'


import herculens.MassModel.profile_mapping as pm
from herculens.Util import util


__all__ = ['MassModelBase']


SUPPORTED_MODELS = pm.SUPPORTED_MODELS
STRING_MAPPING = pm.STRING_MAPPING


class MassModelBase(object):
    """Base class for managing lens models in single- or multi-plane lensing."""
    def __init__(self, profile_list, kwargs_pixelated=None, 
                 **profile_specific_kwargs):
        """Create a MassProfileBase object.

        NOTE: the extra keyword arguments are given to the corresponding profile class
        only when that profile is given as a string instead of a class instance.

        Parameters
        ----------
        profile_list : list of strings or profile instances
            List of mass profiles. If not a list, wrap the passed argument in a list. 
            It can also be None, in which case the model has no mass profile associated to it.
        kwargs_pixelated : dict
            Settings related to the creation of the pixelated grid.
            See herculens.PixelGrid.create_model_grid for details.
        profile_specific_kwargs : dict
            See docstring for get_class_from_string().
        """
        if profile_list is None:
            self.func_list = []
            self._pix_idx = None
            self._num_func = 0
            self._model_list = []
        elif not isinstance(profile_list, (list, tuple)):
            raise TypeError("The profile list should be a list or a tuple.")
        self.func_list, self._pix_idx = self._load_model_instances(
            profile_list, **profile_specific_kwargs,
        )
        self._num_func = len(self.func_list)
        self._model_list = profile_list
        if kwargs_pixelated is None:
            kwargs_pixelated = {}
        self._kwargs_pixelated = kwargs_pixelated
        
    def _load_model_instances(
            self, profile_list, **profile_specific_kwargs,
        ):
        func_list = []
        pix_idx = None
        for idx, profile_type in enumerate(profile_list):
            if isinstance(profile_type, str):
                # passing string is supported for backward-compatibility only
                profile_class = self.get_class_from_string(
                    profile_type, 
                    **profile_specific_kwargs,
                )
                if profile_type in ['PIXELATED', 'PIXELATED_DIRAC']:
                    pix_idx = idx

            # this is the new preferred way: passing the profile as a class
            elif self.is_mass_profile_class(profile_type):
                profile_class = profile_type
                if isinstance(
                        profile_class, 
                        (STRING_MAPPING['PIXELATED'], STRING_MAPPING['PIXELATED_DIRAC'])
                    ):
                    pix_idx = idx
            else:
                raise TypeError(f"Each profile can either be a string or "
                                 f"directly a profile instance (supported profiles are {SUPPORTED_MODELS}).")
            func_list.append(profile_class)
        return func_list, pix_idx
    
    @staticmethod
    def is_mass_profile_class(profile):
        """Simply checks that the mass profile has the required methods"""
        return (
            hasattr(profile, 'function') and
            hasattr(profile, 'derivatives') and
            hasattr(profile, 'hessian')
        )

    @staticmethod
    def get_class_from_string(
            profile_string, 
            pixel_derivative_type=None, 
            pixel_interpol=None, 
            no_complex_numbers=None, 
            kwargs_pixel_grid_fixed=None,
        ):
        """
        Get the lens profile class of the corresponding type.
        Keyword arguments are related to specific profile types.
        
        Parameters
        ----------
        smoothing : float
            Smoothing factor for some models (deprecated).
        pixel_interpol : string
            Type of interpolation for 'PIXELATED' profiles: 'fast_bilinear' or 'bicubic'
        pixel_derivative_type : str
            Type of interpolation: 'interpol' or 'autodiff'
        no_complex_numbers : bool
            Use or not complex number in the EPL's deflection computation.
        kwargs_pixel_grid_fixed : dict
            Settings related to the creation of the pixelated grid for profile type 'PIXELATED_FIXED'.
            See herculens.PixelGrid.create_model_grid for details.
        """
        if profile_string in SUPPORTED_MODELS:
            profile_class = STRING_MAPPING[profile_string]
            kwargs = {}
            # treats the few special cases that require user settings
            if profile_string == 'EPL':
                if no_complex_numbers is not None:
                    kwargs['no_complex_numbers'] = no_complex_numbers
                return profile_class(**kwargs)
            elif profile_string == 'PIXELATED':
                if pixel_interpol is None:
                    kwargs['interpolation_type'] = pixel_interpol
                if pixel_derivative_type is None:
                    kwargs['derivative_type'] = pixel_derivative_type
                return profile_class(**kwargs)
            elif profile_string == 'PIXELATED_FIXED':
                if kwargs_pixel_grid_fixed is None:
                    raise ValueError("At least one pixel grid must be provided to use 'PIXELATED_FIXED' profile")
                return profile_class(**kwargs_pixel_grid_fixed)
        else:
            raise ValueError(f"Could not load profile type '{profile_string}'.")
        # all remaining profiles take no extra arguments
        return profile_class()

    def _bool_list(self, k):
        return util.convert_bool_list(n=self._num_func, k=k)

    @property
    def has_pixels(self):
        return self._pix_idx is not None

    @property
    def pixel_grid_settings(self):
        return self._kwargs_pixelated

    def set_pixel_grid(self, pixel_grid):
        self.func_list[self.pixelated_index].set_pixel_grid(pixel_grid)

    @property
    def pixel_grid(self):
        if not self.has_pixels:
            return None
        return self.func_list[self.pixelated_index].pixel_grid

    @property
    def pixelated_index(self):
        # TODO: support multiple pixelated profiles
        return self._pix_idx

    @property
    def pixelated_coordinates(self):
        if not self.has_pixels:
            return None, None
        return self.pixel_grid.pixel_coordinates

    @property
    def pixelated_shape(self):
        if not self.has_pixels:
            return None
        x_coords, _ = self.pixelated_coordinates
        return x_coords.shape



# Defines Sersic profiles
# 
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the LightModel.Profiles module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'jiwoncpark', 'austinpeel', 'aymgal'


import numpy as np
import jax.numpy as jnp
from jax import jit, grad
from functools import partial

from herculens.MassModel.Profiles.sersic_utils import SersicUtil
import herculens.Util.param_util as param_util


__all__ = ['Sersic', 'SersicElliptic']


class Sersic(SersicUtil):
    """
    this class contains functions to evaluate an spherical Sersic function

    .. math::

        I(R) = I_0 \exp \left[ -b_n (R/R_{\rm Sersic})^{\frac{1}{n}}\right]

    with :math:`I_0 = amp`
    and
    with :math:`b_{n}\approx 1.999\,n-0.327`

    """

    param_names = ['amp', 'R_sersic', 'n_sersic', 'center_x', 'center_y']
    lower_limit_default = {'amp': 0, 'R_sersic': 0, 'n_sersic': 0.5, 'center_x': -100, 'center_y': -100}
    upper_limit_default = {'amp': 100, 'R_sersic': 100, 'n_sersic': 8, 'center_x': 100, 'center_y': 100}
    fixed_default = {key: False for key in param_names}

    def __init__(self, smoothing=0.00001, exponent=2.):
        super().__init__(smoothing=smoothing, exponent=exponent)

    def function(self, x, y, R_sersic, n_sersic, center_x=0, center_y=0, amp=1., max_R_frac=100.0):
        """

        :param x:
        :param y:
        :param amp: surface brightness/amplitude value at the half light radius
        :param R_sersic: semi-major axis half light radius
        :param n_sersic: Sersic index
        :param center_x: center in x-coordinate
        :param center_y: center in y-coordinate
        :param max_R_frac: maximum window outside of which the mass is zeroed, in units of R_sersic (float)
        :return: Sersic profile value at (x, y)
        """
        R = self.get_distance_from_center(x, y, phi_G=0.0, q=1.0, center_x=center_x, center_y=center_y)
        result = self._r_sersic(R, R_sersic, n_sersic, max_R_frac)
        return amp * result

    def derivatives(self, x, y, R_sersic, n_sersic, center_x, center_y, amp, max_R_frac=100.0):
        """

        :param x:
        :param y:
        :param amp: surface brightness/amplitude value at the half light radius
        :param R_sersic: semi-major axis half light radius
        :param n_sersic: Sersic index
        :param e1: eccentricity parameter
        :param e2: eccentricity parameter
        :param center_x: center in x-coordinate
        :param center_y: center in y-coordinate
        :param max_R_frac: maximum window outside of which the mass is zeroed, in units of R_sersic (float)
        :return: partial derivatives of Sersic profile value at (x, y) with respect to x and y
        """
        def _function(p):
            return self.function(p[0], p[1], 
                                 R_sersic, n_sersic,
                                 center_x, center_y, amp,
                                 max_R_frac=max_R_frac)
        
        grad_function = grad(_function)

        @jit
        def _grad_function(x, y):
            return grad_function([x, y])[0], grad_function([x, y])[1]
        
        f_x, f_y = jnp.vectorize(_grad_function)(x, y)
        return f_x, f_y

    def derivatives_explicit(self, x, y, R_sersic, n_sersic, center_x=0, center_y=0, amp=1., max_R_frac=100.0):
        """

        :param x:
        :param y:
        :param amp: surface brightness/amplitude value at the half light radius
        :param R_sersic: semi-major axis half light radius
        :param n_sersic: Sersic index
        :param e1: eccentricity parameter
        :param e2: eccentricity parameter
        :param center_x: center in x-coordinate
        :param center_y: center in y-coordinate
        :param max_R_frac: maximum window outside of which the mass is zeroed, in units of R_sersic (float)
        :return: partial derivatives of Sersic profile value at (x, y) with respect to x and y
        """
        x_ = np.array(x) - center_x
        y_ = np.array(y) - center_y
        r = np.sqrt(x_**2 + y_**2)
        #if isinstance(r, int) or isinstance(r, float):
        #    r = max(self._s, r)
        #else:
        #    r[r < self._s] = self._s
        alpha = -self.alpha_abs(x, y, n_sersic, R_sersic, amp, center_x, center_y)
        f_x = alpha * x_ / r
        f_y = alpha * y_ / r
        return f_x, f_y

    @property
    def num_amplitudes(self):
        return 1


class SersicElliptic(SersicUtil):
    """
    this class contains functions to evaluate an elliptical Sersic function
    """
    param_names = ['amp', 'R_sersic', 'n_sersic', 'e1', 'e2', 'center_x', 'center_y']
    lower_limit_default = {'amp': 0, 'R_sersic': 0, 'n_sersic': 0.5, 'e1': -0.5, 'e2': -0.5,'center_x': -100, 'center_y': -100}
    upper_limit_default = {'amp': 100, 'R_sersic': 100, 'n_sersic': 8, 'e1': 0.5, 'e2': 0.5,'center_x': 100, 'center_y': 100}
    fixed_default = {key: False for key in param_names}

    def __init__(self, smoothing=0.00001, exponent=2.):
        super().__init__(smoothing=smoothing, exponent=exponent)

    def function(self, x, y, R_sersic, n_sersic, e1, e2, center_x, center_y, amp, max_R_frac=1000.0):
        """

        :param x:
        :param y:
        :param amp: surface brightness/amplitude value at the half light radius
        :param R_sersic: semi-major axis half light radius
        :param n_sersic: Sersic index
        :param e1: eccentricity parameter
        :param e2: eccentricity parameter
        :param center_x: center in x-coordinate
        :param center_y: center in y-coordinate
        :param max_R_frac: maximum window outside of which the mass is zeroed, in units of R_sersic (float)
        :return: Sersic profile value at (x, y)
        """

        R_sersic = jnp.maximum(0, R_sersic)
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        R = self.get_distance_from_center(x, y, phi_G, q, center_x, center_y)
        result = self._r_sersic(R, R_sersic, n_sersic, max_R_frac)
        return amp * result

    def derivatives(self, x, y, R_sersic, n_sersic, e1, e2, center_x, center_y, amp, max_R_frac=1000.0):
        """

        :param x:
        :param y:
        :param amp: surface brightness/amplitude value at the half light radius
        :param R_sersic: semi-major axis half light radius
        :param n_sersic: Sersic index
        :param e1: eccentricity parameter
        :param e2: eccentricity parameter
        :param center_x: center in x-coordinate
        :param center_y: center in y-coordinate
        :param max_R_frac: maximum window outside of which the mass is zeroed, in units of R_sersic (float)
        :return: partial derivatives of Sersic profile value at (x, y) with respect to x and y
        """
        def _function(p):
            return self.function(p[0], p[1], 
                                 R_sersic, n_sersic, e1, e2,
                                 center_x, center_y, amp,
                                 max_R_frac=max_R_frac)
        
        grad_function = grad(_function)

        @jit
        def _grad_function(x, y):
            return grad_function([x, y])[0], grad_function([x, y])[1]
        
        f_x, f_y = jnp.vectorize(_grad_function)(x, y)
        return f_x, f_y

    @property
    def num_amplitudes(self):
        return 1



# Utility functions
# 
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the Util module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'austinpeel', 'aymgal'


import numpy as np
import jax.numpy as jnp
from jax import lax


def phi_q2_ellipticity(phi, q):
    """
    transforms orientation angle and axis ratio into complex ellipticity moduli e1, e2

    :param phi: angle of orientation (in radian)
    :param q: axis ratio minor axis / major axis
    :return: eccentricities e1 and e2 in complex ellipticity moduli
    """
    # e1 = jnp.where(q == 1., 1e-8, (1. - q) / (1. + q) * jnp.cos(2. * phi))
    # e2 = jnp.where(q == 1., 1e-8, (1. - q) / (1. + q) * jnp.sin(2. * phi))
    e1 = (1. - q) / (1. + q) * jnp.cos(2. * phi)
    e2 = (1. - q) / (1. + q) * jnp.sin(2. * phi)
    return e1, e2

def ellipticity2phi_q(e1, e2):
    """Transform complex ellipticity components to position angle and axis ratio.

    Parameters
    ----------
    e1, e2 : float or array_like
        Ellipticity components.

    Returns
    -------
    phi, q : same type as e1, e2
        Position angle (rad) and axis ratio (semi-minor / semi-major axis)

    """
    # replace value by low float instead to avoid NaNs
    # e1 = lax.cond(e1 == 0.0, lambda _: 1e-4, lambda _: e1, operand=None)  # does not work with TFP!
    # e2 = lax.cond(e2 == 0.0, lambda _: 1e-4, lambda _: e2, operand=None)  # does not work with TFP!
    # e1 = jnp.where(e1 == 0., 1e-8, e1)
    # e2 = jnp.where(e2 == 0., 1e-8, e2)
    phi = jnp.arctan2(e2, e1) / 2.
    c = jnp.sqrt(e1**2 + e2**2)
    c = jnp.minimum(c, 0.9999)
    q = (1. - c) / (1. + c)
    return phi, q

def ellipticity2phi_q_numpy(e1, e2):
    """Transform complex ellipticity components to position angle and axis ratio.

    Parameters
    ----------
    e1, e2 : float or array_like
        Ellipticity components.

    Returns
    -------
    phi, q : same type as e1, e2
        Position angle (rad) and axis ratio (semi-minor / semi-major axis)

    """
    # replace value by low float instead to avoid NaNs
    e1_ = np.copy(e1)
    e1_[e1 == 0.] == 1e-4
    e2_ = np.copy(e2)
    e2_[e2 == 0.] == 1e-4
    phi = np.arctan2(e2_, e1_) / 2
    c = np.sqrt(e1_**2 + e2_**2)
    c = np.minimum(c, 0.9999)
    q = (1. - c) / (1. + c)
    return phi, q

def shear_polar2cartesian(phi, gamma):
    """

    :param phi: shear angle (radian)
    :param gamma: shear strength
    :return: shear components gamma1, gamma2
    """
    gamma1 = gamma*jnp.cos(2*phi)
    gamma2 = gamma*jnp.sin(2*phi)
    return gamma1, gamma2

def shear_cartesian2polar(gamma1, gamma2):
    """
    :param gamma1: cartesian shear component
    :param gamma2: cartesian shear component
    :return: shear angle, shear strength
    """
    phi = jnp.arctan2(gamma2, gamma1) / 2
    gamma = jnp.sqrt(gamma1 ** 2 + gamma2 ** 2)
    return phi, gamma

def shear_cartesian2polar_numpy(gamma1, gamma2):
    """
    :param gamma1: cartesian shear component
    :param gamma2: cartesian shear component
    :return: shear angle, shear strength
    """
    gamma1 = np.array(gamma1)
    gamma2 = np.array(gamma2)
    phi = np.arctan2(gamma2, gamma1) / 2
    gamma = np.sqrt(gamma1 ** 2 + gamma2 ** 2)
    return phi, gamma
    
def cart2polar(x, y, center_x=0, center_y=0):
    """
    transforms cartesian coords [x,y] into polar coords [r,phi] in the frame of the lens center

    :param x: set of x-coordinates
    :type x: array of size (n)
    :param y: set of x-coordinates
    :type y: array of size (n)
    :param center_x: rotation point
    :type center_x: float
    :param center_y: rotation point
    :type center_y: float
    :returns:  array of same size with coords [r,phi]
    """
    coord_shift_x = x - center_x
    coord_shift_y = y - center_y
    r = jnp.sqrt(coord_shift_x**2+coord_shift_y**2)
    phi = jnp.arctan2(coord_shift_y, coord_shift_x)
    return r, phi

def polar2cart(r, phi, center):
    """
    transforms polar coords [r,phi] into cartesian coords [x,y] in the frame of the lense center

    :param coord: set of coordinates
    :type coord: array of size (n,2)
    :param center: rotation point
    :type center: array of size (2)
    :returns:  array of same size with coords [x,y]
    :raises: AttributeError, KeyError
    """
    x = r*jnp.cos(phi)
    y = r*jnp.sin(phi)
    return x - center[0], y - center[1]

def transform_e1e2_product_average(x, y, e1, e2, center_x, center_y):
    """
    maps the coordinates x, y with eccentricities e1 e2 into a new elliptical coordinate system
    such that R = sqrt(R_major * R_minor)

    :param x: x-coordinate
    :param y: y-coordinate
    :param e1: eccentricity
    :param e2: eccentricity
    :param center_x: center of distortion
    :param center_y: center of distortion
    :return: distorted coordinates x', y'
    """
    phi_G, q = ellipticity2phi_q(e1, e2)
    x_shift = x - center_x
    y_shift = y - center_y

    cos_phi = jnp.cos(phi_G)
    sin_phi = jnp.sin(phi_G)

    xt1 = cos_phi * x_shift + sin_phi * y_shift
    xt2 = -sin_phi * x_shift + cos_phi * y_shift
    return xt1 * jnp.sqrt(q), xt2 / jnp.sqrt(q)

def transform_e1e2_square_average(x, y, e1, e2, center_x, center_y):
    """
    maps the coordinates x, y with eccentricities e1 e2 into a new elliptical coordinate system
    such that R = sqrt(R_major**2 + R_minor**2)

    :param x: x-coordinate
    :param y: y-coordinate
    :param e1: eccentricity
    :param e2: eccentricity
    :param center_x: center of distortion
    :param center_y: center of distortion
    :return: distorted coordinates x', y'
    """
    phi_G, q = ellipticity2phi_q(e1, e2)
    x_shift = x - center_x
    y_shift = y - center_y
    cos_phi = jnp.cos(phi_G)
    sin_phi = jnp.sin(phi_G)
    e = q2e(q)
    x_ = (cos_phi * x_shift + sin_phi * y_shift) * jnp.sqrt(1 - e)
    y_ = (-sin_phi * x_shift + cos_phi * y_shift) * jnp.sqrt(1 + e)
    return x_, y_


def q2e(q):
    e = jnp.abs(1 - q**2) / (1 + q**2)
    return e


def statistics_from_samples(samples, losses):
    min_loss_idx = np.argmin(losses)
    map_values = samples[min_loss_idx, :]
    mean_values = np.mean(samples, axis=0)
    perc_16, perc_50, perc_84 = np.percentile(samples, q=[16, 50, 84], axis=0)
    median_values = perc_50
    return map_values, mean_values, median_values, perc_16, perc_84
